"""Google Calendar skills — reliable workflows via Google Calendar Web.

Strategy: Google Calendar has well-structured DOM elements that Playwright
can interact with reliably. We use the quick-add feature and direct URL
parameters for creating events.

URL scheme: https://calendar.google.com/calendar/r/eventedit?text=TITLE&dates=START/END&details=DETAILS
Date format: YYYYMMDDTHHmmSS (e.g., 20260221T140000)
"""

from __future__ import annotations
from typing import Any
from urllib.parse import quote
from plutus.skills.engine import SkillDefinition, SkillStep


class GoogleCalendarCreateEvent(SkillDefinition):
    name = "calendar_create_event"
    description = "Create a new Google Calendar event with title, date, time, and optional details"
    app = "Google Calendar"
    triggers = ["calendar event", "schedule meeting", "add to calendar", "create event",
                "book meeting", "calendar entry", "reminder", "schedule appointment"]
    category = "calendar"
    required_params = ["title"]
    optional_params = ["date", "start_time", "end_time", "description", "location"]

    def build_steps(self, params: dict[str, Any]) -> list[SkillStep]:
        title = params["title"]
        date = params.get("date", "")  # e.g., "2026-02-21" or "tomorrow"
        start_time = params.get("start_time", "")  # e.g., "14:00"
        end_time = params.get("end_time", "")  # e.g., "15:00"
        description = params.get("description", "")
        location = params.get("location", "")

        # Build the Google Calendar URL with pre-filled fields
        url_parts = [f"https://calendar.google.com/calendar/r/eventedit?text={quote(title)}"]

        if date and start_time:
            # Convert to Google Calendar date format
            # The LLM should provide dates in YYYYMMDD format or we handle common formats
            date_clean = date.replace("-", "")
            start_clean = start_time.replace(":", "") + "00"
            if end_time:
                end_clean = end_time.replace(":", "") + "00"
            else:
                # Default: 1 hour event
                end_clean = start_clean  # will be handled by Google
            url_parts.append(f"dates={date_clean}T{start_clean}/{date_clean}T{end_clean}")

        if description:
            url_parts.append(f"details={quote(description)}")
        if location:
            url_parts.append(f"location={quote(location)}")

        url = "&".join(url_parts)

        steps = [
            SkillStep(
                description=f"Open Google Calendar event creator for: {title}",
                operation="open_url",
                params={"url": url},
                wait_after=4.0,
            ),
            SkillStep(
                description="Wait for the event editor to load",
                operation="wait_for_text",
                params={"text": "Save", "timeout": 10000},
                wait_after=1.0,
                optional=True,
            ),
        ]

        # If no date/time was provided via URL, we need to fill them in manually
        if not date or not start_time:
            steps.append(SkillStep(
                description="Get the event form to check what needs to be filled",
                operation="get_page",
                params={},
                wait_after=0.5,
            ))

        steps.append(SkillStep(
            description="Click Save to create the event",
            operation="browser_click",
            params={"text": "Save"},
            wait_after=2.0,
            retry_on_fail=True,
            max_retries=2,
        ))

        return steps


class GoogleCalendarCheckSchedule(SkillDefinition):
    name = "calendar_check_schedule"
    description = "Check your Google Calendar schedule for today or a specific date"
    app = "Google Calendar"
    triggers = ["check calendar", "what's on my calendar", "schedule today", "my schedule",
                "calendar for today", "upcoming events", "what meetings"]
    category = "calendar"
    required_params = []
    optional_params = ["date"]

    def build_steps(self, params: dict[str, Any]) -> list[SkillStep]:
        date = params.get("date", "")

        if date:
            # Navigate to specific date
            date_clean = date.replace("-", "/")
            url = f"https://calendar.google.com/calendar/r/day/{date_clean}"
        else:
            # Today's schedule
            url = "https://calendar.google.com/calendar/r/day"

        return [
            SkillStep(
                description="Open Google Calendar day view",
                operation="open_url",
                params={"url": url},
                wait_after=4.0,
            ),
            SkillStep(
                description="Wait for calendar to load",
                operation="wait_for_text",
                params={"text": "Day", "timeout": 10000},
                wait_after=1.0,
                optional=True,
            ),
            SkillStep(
                description="Read the calendar content",
                operation="get_page",
                params={},
                wait_after=0.0,
            ),
        ]


class GoogleCalendarDeleteEvent(SkillDefinition):
    name = "calendar_delete_event"
    description = "Delete a Google Calendar event by clicking on it and selecting delete"
    app = "Google Calendar"
    triggers = ["delete event", "remove from calendar", "cancel meeting", "delete calendar entry"]
    category = "calendar"
    required_params = ["event_title"]
    optional_params = ["date"]

    def build_steps(self, params: dict[str, Any]) -> list[SkillStep]:
        event_title = params["event_title"]
        date = params.get("date", "")

        steps = []

        if date:
            date_clean = date.replace("-", "/")
            url = f"https://calendar.google.com/calendar/r/day/{date_clean}"
        else:
            url = "https://calendar.google.com/calendar/r/day"

        steps.extend([
            SkillStep(
                description="Open Google Calendar",
                operation="open_url",
                params={"url": url},
                wait_after=4.0,
            ),
            SkillStep(
                description="Wait for calendar to load",
                operation="wait_for_text",
                params={"text": "Day", "timeout": 10000},
                wait_after=1.0,
                optional=True,
            ),
            SkillStep(
                description=f"Click on the event: {event_title}",
                operation="browser_click",
                params={"text": event_title},
                wait_after=2.0,
                retry_on_fail=True,
            ),
            SkillStep(
                description="Click the delete button (trash icon)",
                operation="browser_click",
                params={"selector": "[aria-label='Delete event']"},
                wait_after=1.0,
                retry_on_fail=True,
                optional=True,
            ),
            # Fallback: look for "Delete" text
            SkillStep(
                description="Click Delete (fallback by text)",
                operation="browser_click",
                params={"text": "Delete"},
                wait_after=1.0,
                optional=True,
            ),
        ])

        return steps



"""Tests for the tool system."""

import os
import tempfile
from pathlib import Path

import pytest

from plutus.tools.base import Tool  # noqa: F401
from plutus.tools.filesystem import FilesystemTool
from plutus.tools.registry import ToolRegistry
from plutus.tools.shell import BLOCKED_COMMANDS, ShellTool  # noqa: F401
from plutus.tools.system_info import SystemInfoTool
from plutus.tools.wsl import BLOCKED_PATTERNS as WSL_BLOCKED_PATTERNS
from plutus.tools.wsl import WSLTool

# ── Tool Registry tests ─────────────────────────────────────


class TestToolRegistry:
    def test_register_and_get(self):
        registry = ToolRegistry()
        tool = ShellTool()
        registry.register(tool)

        assert registry.get("shell") is tool
        assert registry.get("nonexistent") is None

    def test_list_tools(self):
        registry = ToolRegistry()
        registry.register(ShellTool())
        registry.register(FilesystemTool())

        tools = registry.list_tools()
        assert "shell" in tools
        assert "filesystem" in tools

    def test_get_definitions(self):
        registry = ToolRegistry()
        registry.register(ShellTool())
        registry.register(SystemInfoTool())

        defs = registry.get_definitions()
        assert len(defs) == 2
        names = {d.name for d in defs}
        assert names == {"shell", "system_info"}

    def test_get_tool_info(self):
        registry = ToolRegistry()
        registry.register(ShellTool())

        info = registry.get_tool_info()
        assert len(info) == 1
        assert info[0]["name"] == "shell"
        assert "description" in info[0]
        assert "parameters" in info[0]


# ── Shell Tool tests ────────────────────────────────────────


class TestShellTool:
    @pytest.fixture
    def tool(self):
        return ShellTool()

    def test_properties(self, tool):
        assert tool.name == "shell"
        assert len(tool.description) > 0
        assert tool.parameters["type"] == "object"
        assert "command" in tool.parameters["properties"]

    @pytest.mark.asyncio
    async def test_execute_simple_command(self, tool):
        result = await tool.execute(command="echo hello")
        assert "hello" in result
        assert "exit_code: 0" in result

    @pytest.mark.asyncio
    async def test_execute_with_working_directory(self, tool):
        result = await tool.execute(command="pwd", working_directory="/tmp")
        assert "/tmp" in result

    @pytest.mark.asyncio
    async def test_blocks_dangerous_commands(self, tool):
        result = await tool.execute(command="rm -rf /")
        assert "[BLOCKED]" in result

    @pytest.mark.asyncio
    async def test_captures_stderr(self, tool):
        result = await tool.execute(command="ls /nonexistent_path_12345")
        assert "stderr:" in result or "exit_code:" in result

    @pytest.mark.asyncio
    async def test_timeout(self, tool):
        result = await tool.execute(command="sleep 5", timeout=1)
        assert "[TIMEOUT]" in result

    def test_get_definition(self, tool):
        defn = tool.get_definition()
        assert defn.name == "shell"
        assert "command" in defn.parameters["properties"]


# ── Filesystem Tool tests ───────────────────────────────────


class TestFilesystemTool:
    @pytest.fixture
    def tool(self):
        return FilesystemTool()

    @pytest.mark.asyncio
    async def test_read_file(self, tool):
        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as f:
            f.write("test content")
            path = f.name

        result = await tool.execute(operation="read", path=path)
        assert "test content" in result
        os.unlink(path)

    @pytest.mark.asyncio
    async def test_write_file(self, tool):
        with tempfile.NamedTemporaryFile(suffix=".txt", delete=False) as f:
            path = f.name

        result = await tool.execute(
            operation="write", path=path, content="new content"
        )
        assert "written" in result.lower()
        assert Path(path).read_text() == "new content"
        os.unlink(path)

    @pytest.mark.asyncio
    async def test_list_directory(self, tool):
        with tempfile.TemporaryDirectory() as tmpdir:
            Path(tmpdir, "a.txt").write_text("a")
            Path(tmpdir, "b.txt").write_text("b")

            result = await tool.execute(operation="list", path=tmpdir)
            assert "a.txt" in result
            assert "b.txt" in result

    @pytest.mark.asyncio
    async def test_mkdir(self, tool):
        with tempfile.TemporaryDirectory() as tmpdir:
            new_dir = os.path.join(tmpdir, "subdir", "nested")
            result = await tool.execute(operation="mkdir", path=new_dir)
            assert "Created" in result
            assert Path(new_dir).is_dir()

    @pytest.mark.asyncio
    async def test_info(self, tool):
        with tempfile.NamedTemporaryFile(suffix=".txt", delete=False) as f:
            f.write(b"test")
            path = f.name

        result = await tool.execute(operation="info", path=path)
        assert "file" in result
        assert "size:" in result
        os.unlink(path)

    @pytest.mark.asyncio
    async def test_read_nonexistent(self, tool):
        result = await tool.execute(operation="read", path="/nonexistent_12345.txt")
        assert "[ERROR]" in result

    @pytest.mark.asyncio
    async def test_search(self, tool):
        with tempfile.TemporaryDirectory() as tmpdir:
            Path(tmpdir, "test.py").write_text("def hello_world():\n    pass\n")
            Path(tmpdir, "other.txt").write_text("nothing here\n")

            result = await tool.execute(
                operation="search", path=tmpdir, pattern="hello_world"
            )
            assert "hello_world" in result
            assert "test.py" in result


# ── SystemInfo Tool tests ───────────────────────────────────


class TestSystemInfoTool:
    @pytest.fixture
    def tool(self):
        return SystemInfoTool()

    @pytest.mark.asyncio
    async def test_overview(self, tool):
        result = await tool.execute(operation="overview")
        assert "CPU:" in result
        assert "Memory:" in result

    @pytest.mark.asyncio
    async def test_os_info(self, tool):
        result = await tool.execute(operation="os")
        assert "System:" in result
        assert "Python:" in result


# ── WSL Tool tests ─────────────────────────────────────────


class TestWSLTool:
    @pytest.fixture
    def tool(self):
        return WSLTool()

    def test_properties(self, tool):
        assert tool.name == "wsl"
        assert len(tool.description) > 0
        assert tool.parameters["type"] == "object"
        assert "operation" in tool.parameters["properties"]
        assert "command" in tool.parameters["properties"]

    def test_get_definition(self, tool):
        defn = tool.get_definition()
        assert defn.name == "wsl"
        assert "operation" in defn.parameters["properties"]

    @pytest.mark.asyncio
    async def test_run_simple_command(self, tool):
        """On Linux (CI), the WSL tool falls back to native shell."""
        result = await tool.execute(operation="run", command="echo wsl_test")
        assert "wsl_test" in result
        assert "exit_code: 0" in result

    @pytest.mark.asyncio
    async def test_run_requires_command(self, tool):
        result = await tool.execute(operation="run")
        assert "[ERROR]" in result

    @pytest.mark.asyncio
    async def test_blocks_dangerous_commands(self, tool):
        result = await tool.execute(operation="run", command="rm -rf /")
        assert "[BLOCKED]" in result

    @pytest.mark.asyncio
    async def test_blocks_fork_bomb(self, tool):
        result = await tool.execute(operation="run", command=":(){ :|:& };:")
        assert "[BLOCKED]" in result

    @pytest.mark.asyncio
    async def test_timeout(self, tool):
        result = await tool.execute(operation="run", command="sleep 5", timeout=1)
        assert "[TIMEOUT]" in result

    @pytest.mark.asyncio
    async def test_unknown_operation(self, tool):
        result = await tool.execute(operation="bogus")
        assert "[ERROR]" in result

    @pytest.mark.asyncio
    async def test_info(self, tool):
        result = await tool.execute(operation="info")
        assert "host_os:" in result

    @pytest.mark.asyncio
    async def test_list_distros_non_windows(self, tool):
        """On Linux/macOS, list_distros returns a helpful non-applicable message."""
        import platform
        if platform.system() == "Windows":
            pytest.skip("Test only for non-Windows")
        result = await tool.execute(operation="list_distros")
        assert "native" in result.lower() or "not applicable" in result.lower()

    @pytest.mark.asyncio
    async def test_path_to_linux_non_windows(self, tool):
        import platform
        if platform.system() == "Windows":
            pytest.skip("Test only for non-Windows")
        result = await tool.execute(operation="path_to_linux", path="/tmp/test.txt")
        assert "/tmp/test.txt" in result

    @pytest.mark.asyncio
    async def test_path_requires_path(self, tool):
        result = await tool.execute(operation="path_to_linux")
        assert "[ERROR]" in result

    @pytest.mark.asyncio
    async def test_set_default_non_windows(self, tool):
        import platform
        if platform.system() == "Windows":
            pytest.skip("Test only for non-Windows")
        result = await tool.execute(operation="set_default", distro="Ubuntu")
        assert "[ERROR]" in result

    @pytest.mark.asyncio
    async def test_working_directory(self, tool):
        result = await tool.execute(
            operation="run", command="pwd", working_directory="/tmp"
        )
        assert "/tmp" in result

    def test_blocked_patterns_exist(self):
        assert len(WSL_BLOCKED_PATTERNS) > 0
        assert "rm -rf /" in WSL_BLOCKED_PATTERNS



"""Custom hatch build hook to generate platform-specific wheels.

When compiled Cython extensions (.so/.pyd) are present, hatchling must
produce a platform wheel (e.g. cp311-cp311-macosx_11_0_arm64) instead
of a pure-python wheel (py3-none-any). This hook tells hatchling to
infer the platform tag from the current Python interpreter.
"""

from __future__ import annotations

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomBuildHook(BuildHookInterface):
    def initialize(self, version, build_data):
        # Force platform-specific wheel — required because we include
        # pre-compiled Cython .so/.pyd extensions via artifacts
        build_data["pure_python"] = False
        build_data["infer_tag"] = True



"""Plutus PC Control — four-layer computer interaction.

Architecture (matching OpenClaw's proven approach):
  Layer 1: OS Control (shell commands — most reliable)
  Layer 2: Browser Control (Playwright accessibility tree — for web interaction)
  Layer 2.5: Desktop UIA (pywinauto accessibility tree — for native Windows apps)
  Layer 3: Desktop Fallback (PyAutoGUI — last resort for native apps)

The OSControl is the PRIMARY interface for opening apps, running commands,
and managing processes via native OS commands.

The BrowserControl handles all web interaction via Playwright with
accessibility tree snapshots and ref-based interaction.

The DesktopControl handles native Windows app interaction via pywinauto
with UIA accessibility tree snapshots and ref-based interaction.

Desktop fallback (mouse, keyboard, screen) is a last resort.
"""

from plutus.pc.os_control import OSControl
from plutus.pc.browser_control import BrowserControl
from plutus.pc.desktop_control import DesktopControl
from plutus.pc.mouse import MouseController
from plutus.pc.keyboard import KeyboardController
from plutus.pc.screen import ScreenReader
from plutus.pc.windows import WindowManager
from plutus.pc.workflow import WorkflowEngine
from plutus.pc.context import ContextEngine, ActionGuard, get_context_engine
from plutus.pc.computer_use import ComputerUseExecutor

__all__ = [
    "OSControl",
    "BrowserControl",
    "DesktopControl",
    "ComputerUseExecutor",
    "MouseController",
    "KeyboardController",
    "ScreenReader",
    "WindowManager",
    "WorkflowEngine",
    "ContextEngine",
    "ActionGuard",
    "get_context_engine",
]



"""Base skill definition — skills extend the agent's capabilities."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel


class SkillStep(BaseModel):
    """A single step in a skill execution."""

    tool: str = "shell"
    operation: str | None = None
    params: dict[str, Any] = {}
    run: str | None = None  # shorthand for shell commands

    def to_tool_call(self) -> tuple[str, dict[str, Any]]:
        if self.run:
            return "shell", {"command": self.run}
        return self.tool, {"operation": self.operation, **self.params}


class Skill(BaseModel):
    """A reusable skill that the agent can invoke."""

    name: str
    description: str
    tools_required: list[str] = []
    tier_minimum: str = "operator"
    tags: list[str] = []
    steps: list[SkillStep] = []
    prompt: str | None = None  # optional LLM prompt for complex skills

    def validate_tier(self, current_tier_level: int) -> bool:
        """Check if the current tier meets the minimum for this skill."""
        from plutus.guardrails.tiers import Tier

        try:
            min_tier = Tier(self.tier_minimum)
            return current_tier_level >= min_tier.level
        except ValueError:
            return False



"""Tests for the persistent memory system.

Tests cover:
  - MemoryStore: facts, goals, summaries, checkpoints
  - ConversationSummarizer: message formatting, summary parsing
  - ConversationManager: smart context window, summarization trigger
  - MemoryTool: all actions
"""

import asyncio
import json
import os
import tempfile
import time

import pytest
import pytest_asyncio

# ── MemoryStore tests ──────────────────────────────────────────


@pytest_asyncio.fixture
async def memory():
    """Create a temporary MemoryStore for testing."""
    from plutus.core.memory import MemoryStore

    with tempfile.NamedTemporaryFile(suffix=".db", delete=False) as f:
        db_path = f.name

    store = MemoryStore(db_path)
    await store.initialize()
    yield store
    await store.close()
    os.unlink(db_path)


@pytest.mark.asyncio
async def test_facts_crud(memory):
    """Test creating, reading, searching, and deleting facts."""
    # Store a fact
    fact_id = await memory.store_fact("task_context", "User wants a React app", source="agent")
    assert fact_id > 0

    # Retrieve facts
    facts = await memory.get_facts(category="task_context")
    assert len(facts) == 1
    assert facts[0]["content"] == "User wants a React app"
    assert facts[0]["category"] == "task_context"

    # Store duplicate — should update, not create new
    dup_id = await memory.store_fact("task_context", "User wants a React app", source="agent")
    assert dup_id == fact_id  # Same ID

    facts = await memory.get_facts()
    assert len(facts) == 1  # Still just one

    # Search facts
    results = await memory.search_facts("React")
    assert len(results) == 1

    results = await memory.search_facts("Angular")
    assert len(results) == 0

    # Delete fact
    await memory.delete_fact(fact_id)
    facts = await memory.get_facts()
    assert len(facts) == 0


@pytest.mark.asyncio
async def test_goals_crud(memory):
    """Test creating, reading, and updating goals."""
    await memory.create_conversation("conv-1", "Test conversation")

    # Add goals
    g1 = await memory.add_goal("Build REST API", conversation_id="conv-1", priority=10)
    g2 = await memory.add_goal("Add authentication", conversation_id="conv-1", priority=5)
    g3 = await memory.add_goal("Write tests", conversation_id="conv-1", priority=1)

    assert g1 > 0
    assert g2 > 0
    assert g3 > 0

    # Get active goals (should be ordered by priority DESC)
    active = await memory.get_active_goals(conversation_id="conv-1")
    assert len(active) == 3
    assert active[0]["description"] == "Build REST API"
    assert active[0]["priority"] == 10

    # Complete a goal
    await memory.update_goal_status(g1, "completed")
    active = await memory.get_active_goals(conversation_id="conv-1")
    assert len(active) == 2

    # Get all goals (including completed)
    all_goals = await memory.get_all_goals(conversation_id="conv-1")
    assert len(all_goals) == 3


@pytest.mark.asyncio
async def test_conversation_summaries(memory):
    """Test saving and retrieving conversation summaries."""
    await memory.create_conversation("conv-1", "Test")

    summary_data = {
        "goals": ["Build a web app"],
        "progress": ["Set up project"],
        "current_state": "Working on frontend",
        "key_facts": ["Using React + TypeScript"],
        "next_steps": ["Add routing"],
        "summary": "User wants a web app. Project set up with React.",
    }

    await memory.save_conversation_summary("conv-1", summary_data)

    # Retrieve
    retrieved = await memory.get_conversation_summary("conv-1")
    assert retrieved is not None
    assert retrieved["goals"] == ["Build a web app"]
    assert retrieved["current_state"] == "Working on frontend"

    # Update (upsert)
    summary_data["progress"].append("Added frontend components")
    await memory.save_conversation_summary("conv-1", summary_data)

    retrieved = await memory.get_conversation_summary("conv-1")
    assert len(retrieved["progress"]) == 2


@pytest.mark.asyncio
async def test_checkpoints(memory):
    """Test saving and retrieving checkpoints."""
    await memory.create_conversation("conv-1", "Test")

    # Save checkpoints
    cp1 = await memory.save_checkpoint(
        "conv-1",
        {"working_on": "Step 1", "done": [], "next": ["Step 2"]},
        checkpoint_type="auto",
    )
    assert cp1 > 0

    cp2 = await memory.save_checkpoint(
        "conv-1",
        {"working_on": "Step 2", "done": ["Step 1"], "next": ["Step 3"]},
        checkpoint_type="manual",
    )

    # Get latest
    latest = await memory.get_latest_checkpoint("conv-1")
    assert latest is not None
    assert latest["state_data"]["working_on"] == "Step 2"
    assert latest["checkpoint_type"] == "manual"

    # List all
    all_cps = await memory.list_checkpoints("conv-1")
    assert len(all_cps) == 2


@pytest.mark.asyncio
async def test_memory_stats(memory):
    """Test memory statistics."""
    await memory.create_conversation("conv-1", "Test")
    await memory.add_message("conv-1", "user", content="Hello")
    await memory.add_message("conv-1", "assistant", content="Hi there!")
    await memory.store_fact("general", "Test fact")
    await memory.add_goal("Test goal", conversation_id="conv-1")

    stats = await memory.get_memory_stats()
    assert stats["conversations"] == 1
    assert stats["messages"] == 2
    assert stats["facts"] == 1
    assert stats["active_goals"] == 1


# ── Summarizer tests ──────────────────────────────────────────


def test_format_messages_for_summary():
    """Test that messages are correctly formatted for the summarizer."""
    from plutus.core.summarizer import _format_messages_for_summary

    messages = [
        {"role": "system", "content": "You are Plutus"},
        {"role": "user", "content": "Build me a web app"},
        {"role": "assistant", "content": "I'll create a plan first."},
        {"role": "assistant", "content": None, "tool_calls": [
            {"name": "plan", "arguments": {"action": "create", "title": "Build web app"}}
        ]},
        {"role": "tool", "content": "Plan created", "tool_call_id": "tc_1"},
    ]

    text = _format_messages_for_summary(messages)

    # System messages should be skipped
    assert "You are Plutus" not in text
    # User and assistant messages should be included
    assert "[User]: Build me a web app" in text
    assert "[Assistant]: I'll create a plan first." in text
    assert "[Assistant called plan]" in text
    assert "[Tool Result" in text


def test_parse_summary_json():
    """Test parsing a valid JSON summary."""
    from plutus.core.summarizer import _parse_summary

    json_text = json.dumps({
        "goals": ["Build a web app"],
        "progress": ["Set up project"],
        "current_state": "Working on frontend",
        "key_decisions": ["Using React"],
        "key_facts": ["Project at /home/user/app"],
        "blockers": [],
        "next_steps": ["Add routing"],
        "summary": "Building a web app with React.",
    })

    result = _parse_summary(json_text)
    assert result["goals"] == ["Build a web app"]
    assert result["current_state"] == "Working on frontend"
    assert "created_at" in result


def test_parse_summary_markdown_block():
    """Test parsing JSON wrapped in markdown code blocks."""
    from plutus.core.summarizer import _parse_summary

    text = '```json\n{"goals": ["Test"], "progress": [], "current_state": "", "key_decisions": [], "key_facts": [], "blockers": [], "next_steps": [], "summary": "Test"}\n```'

    result = _parse_summary(text)
    assert result["goals"] == ["Test"]


def test_parse_summary_invalid_json():
    """Test fallback when JSON parsing fails."""
    from plutus.core.summarizer import _parse_summary

    result = _parse_summary("This is not JSON at all")
    assert result["summary"] == "This is not JSON at all"
    assert result["goals"] == []


def test_fallback_summary():
    """Test the fallback summary when LLM fails."""
    from plutus.core.summarizer import _fallback_summary

    messages = [
        {"role": "user", "content": "Build me a React dashboard"},
        {"role": "assistant", "content": "Sure, let me start."},
        {"role": "user", "content": "Add charts too"},
    ]

    result = _fallback_summary(messages)
    assert len(result["goals"]) == 1
    assert "React dashboard" in result["goals"][0]
    assert "3 messages" in result["summary"]


def test_format_summary_for_context():
    """Test formatting a summary for injection into the system prompt."""
    from plutus.core.summarizer import ConversationSummarizer

    # We don't need a real LLM client for this method
    summarizer = ConversationSummarizer(llm_client=None)

    summary = {
        "goals": ["Build a web app", "Add user auth"],
        "progress": ["Project set up", "Frontend done"],
        "current_state": "Working on backend API",
        "key_decisions": ["Using FastAPI for backend"],
        "key_facts": ["Project at /home/user/app", "Using PostgreSQL"],
        "blockers": ["Need to install PostgreSQL"],
        "next_steps": ["Create API endpoints", "Add auth middleware"],
        "summary": "Building a full-stack web app with React frontend and FastAPI backend.",
    }

    text = summarizer.format_summary_for_context(summary)

    assert "## Conversation History Summary" in text
    assert "Build a web app" in text
    assert "Working on backend API" in text
    assert "Using FastAPI for backend" in text
    assert "/home/user/app" in text
    assert "Create API endpoints" in text


# ── ConversationManager tests ─────────────────────────────────


@pytest_asyncio.fixture
async def conversation(memory):
    """Create a ConversationManager for testing."""
    from plutus.core.conversation import ConversationManager
    from plutus.core.planner import PlanManager

    planner = PlanManager(memory)
    await planner.initialize()

    # No summarizer for basic tests (would need LLM)
    manager = ConversationManager(
        memory=memory,
        context_window=10,
        planner=planner,
        summarizer=None,
    )
    return manager


@pytest.mark.asyncio
async def test_conversation_lifecycle(conversation):
    """Test starting a conversation, adding messages, and building context."""
    conv_id = await conversation.start_conversation(title="Test Chat")
    assert conv_id is not None
    assert conversation.conversation_id == conv_id

    # Add messages
    await conversation.add_user_message("Hello, build me an app")
    await conversation.add_assistant_message(content="Sure, I'll create a plan.")

    # Build messages
    messages = await conversation.build_messages()

    # Should have: system + user + assistant
    assert len(messages) >= 3
    assert messages[0]["role"] == "system"

    # Find user and assistant messages
    user_msgs = [m for m in messages if m["role"] == "user"]
    assistant_msgs = [m for m in messages if m["role"] == "assistant"]
    assert len(user_msgs) == 1
    assert len(assistant_msgs) == 1


@pytest.mark.asyncio
async def test_context_window_limits(conversation):
    """Test that context window limits are respected."""
    conv_id = await conversation.start_conversation(title="Long Chat")

    # Add more messages than the context window (10)
    for i in range(15):
        await conversation.add_user_message(f"Message {i}")
        await conversation.add_assistant_message(content=f"Response {i}")

    # Build messages — should only have recent ones
    messages = await conversation.build_messages()

    # Count non-system messages
    non_system = [m for m in messages if m["role"] != "system"]
    # Should be at most context_window messages
    assert len(non_system) <= 10


@pytest.mark.asyncio
async def test_plan_injection(conversation):
    """Test that active plans are injected into context."""
    conv_id = await conversation.start_conversation(title="Plan Test")

    # Create a plan via the planner
    await conversation._planner.create_plan(
        title="Build App",
        steps=[
            {"description": "Set up project"},
            {"description": "Create components"},
        ],
        goal="Build a React app",
        conversation_id=conv_id,
    )

    await conversation.add_user_message("What's the plan?")

    messages = await conversation.build_messages()

    # Find the system message with the plan
    plan_messages = [
        m for m in messages
        if m["role"] == "system" and "Active Plan" in (m.get("content") or "")
    ]
    assert len(plan_messages) == 1
    assert "Build App" in plan_messages[0]["content"]


@pytest.mark.asyncio
async def test_facts_injection(conversation):
    """Test that facts are injected into context."""
    conv_id = await conversation.start_conversation(title="Facts Test")

    # Store some facts
    await conversation._memory.store_fact("task_context", "User prefers TypeScript")
    await conversation._memory.store_fact("file_path", "Project at /home/user/app")

    await conversation.add_user_message("Continue working")

    messages = await conversation.build_messages()

    # Find the system message with facts
    fact_messages = [
        m for m in messages
        if m["role"] == "system" and "Known facts" in (m.get("content") or "")
    ]
    assert len(fact_messages) == 1
    assert "TypeScript" in fact_messages[0]["content"]


# ── Summary merging tests ─────────────────────────────────────


def test_merge_summaries():
    """Test that summaries are correctly merged."""
    from plutus.core.conversation import _merge_summaries

    old = {
        "goals": ["Build a web app"],
        "progress": ["Set up project"],
        "key_facts": ["Using React"],
        "key_decisions": ["Chose React over Vue"],
        "current_state": "Old state",
        "next_steps": ["Old next step"],
    }

    new = {
        "goals": ["Build a web app", "Add auth"],
        "progress": ["Added components"],
        "key_facts": ["Using React", "Database is PostgreSQL"],
        "key_decisions": ["Chose FastAPI for backend"],
        "current_state": "New state",
        "next_steps": ["New next step"],
    }

    merged = _merge_summaries(old, new)

    # Goals should be union
    assert "Build a web app" in merged["goals"]
    assert "Add auth" in merged["goals"]

    # Key facts should be union
    assert "Using React" in merged["key_facts"]
    assert "Database is PostgreSQL" in merged["key_facts"]

    # Key decisions should be accumulated
    assert "Chose React over Vue" in merged["key_decisions"]
    assert "Chose FastAPI for backend" in merged["key_decisions"]

    # Current state should be from new
    assert merged["current_state"] == "New state"

    # Progress should include old items
    assert "Set up project" in merged["progress"]
    assert "Added components" in merged["progress"]


# ── MemoryTool tests ──────────────────────────────────────────


@pytest_asyncio.fixture
async def memory_tool(memory, conversation):
    """Create a MemoryTool for testing."""
    from plutus.tools.memory_tool import MemoryTool

    await conversation.start_conversation(title="Memory Tool Test")
    tool = MemoryTool(memory, conversation)
    return tool


@pytest.mark.asyncio
async def test_memory_tool_save_and_recall(memory_tool):
    """Test saving and recalling facts via the memory tool."""
    # Save a fact
    result = await memory_tool.execute(
        action="save_fact",
        category="task_context",
        content="User wants a dashboard with charts",
    )
    data = json.loads(result)
    assert data["saved"] is True
    assert data["fact_id"] > 0

    # Recall facts
    result = await memory_tool.execute(
        action="recall_facts",
        category="task_context",
    )
    data = json.loads(result)
    assert data["count"] == 1
    assert "dashboard" in data["facts"][0]["content"]


@pytest.mark.asyncio
async def test_memory_tool_goals(memory_tool):
    """Test goal management via the memory tool."""
    # Add a goal
    result = await memory_tool.execute(
        action="add_goal",
        goal_description="Build user authentication system",
    )
    data = json.loads(result)
    assert data["created"] is True
    goal_id = data["goal_id"]

    # List goals
    result = await memory_tool.execute(action="list_goals")
    data = json.loads(result)
    assert data["count"] == 1

    # Complete goal
    result = await memory_tool.execute(
        action="complete_goal",
        goal_id=goal_id,
    )
    data = json.loads(result)
    assert data["status"] == "completed"

    # List goals again (should be empty since completed)
    result = await memory_tool.execute(action="list_goals")
    data = json.loads(result)
    assert data["count"] == 0


@pytest.mark.asyncio
async def test_memory_tool_checkpoint(memory_tool):
    """Test checkpoint operations via the memory tool."""
    # Save checkpoint
    result = await memory_tool.execute(
        action="checkpoint",
        checkpoint_data={
            "working_on": "Building API endpoints",
            "done": ["Project setup", "Database schema"],
            "next": ["Add authentication", "Write tests"],
        },
    )
    data = json.loads(result)
    assert data["saved"] is True

    # Get checkpoint
    result = await memory_tool.execute(action="get_checkpoint")
    data = json.loads(result)
    assert data["checkpoint"] is not None
    assert data["checkpoint"]["state_data"]["working_on"] == "Building API endpoints"


@pytest.mark.asyncio
async def test_memory_tool_stats(memory_tool):
    """Test stats via the memory tool."""
    result = await memory_tool.execute(action="stats")
    data = json.loads(result)
    assert "conversations" in data
    assert "messages" in data
    assert "facts" in data


@pytest.mark.asyncio
async def test_memory_tool_search(memory_tool):
    """Test searching facts via the memory tool."""
    # Save some facts
    await memory_tool.execute(
        action="save_fact", category="technical", content="Using PostgreSQL database"
    )
    await memory_tool.execute(
        action="save_fact", category="technical", content="Using Redis for caching"
    )
    await memory_tool.execute(
        action="save_fact", category="decision", content="Chose React for frontend"
    )

    # Search
    result = await memory_tool.execute(action="search_facts", content="PostgreSQL")
    data = json.loads(result)
    assert data["count"] == 1
    assert "PostgreSQL" in data["facts"][0]["content"]


@pytest.mark.asyncio
async def test_memory_tool_error_handling(memory_tool):
    """Test error handling in the memory tool."""
    # Save fact without content
    result = await memory_tool.execute(action="save_fact")
    assert "[ERROR]" in result

    # Add goal without description
    result = await memory_tool.execute(action="add_goal")
    assert "[ERROR]" in result

    # Unknown action
    result = await memory_tool.execute(action="unknown_action")
    assert "[ERROR]" in result


# ── Integration test ──────────────────────────────────────────


@pytest.mark.asyncio
async def test_full_memory_workflow(memory):
    """Test the full memory workflow: conversation → facts → goals → summary → checkpoint."""
    from plutus.core.conversation import ConversationManager
    from plutus.core.planner import PlanManager
    from plutus.tools.memory_tool import MemoryTool

    planner = PlanManager(memory)
    await planner.initialize()

    conversation = ConversationManager(
        memory=memory,
        context_window=10,
        planner=planner,
        summarizer=None,
    )

    conv_id = await conversation.start_conversation(title="Full Workflow Test")
    tool = MemoryTool(memory, conversation)

    # 1. User asks for something
    await conversation.add_user_message("Build me a React dashboard with charts")

    # 2. Agent saves the goal
    result = await tool.execute(
        action="add_goal",
        goal_description="Build React dashboard with charts",
    )
    goal_data = json.loads(result)

    # 3. Agent saves key facts
    await tool.execute(
        action="save_fact",
        category="task_context",
        content="User wants a React dashboard with charts",
    )
    await tool.execute(
        action="save_fact",
        category="decision",
        content="Using Chart.js for charts",
    )

    # 4. Agent creates a plan
    plan = await planner.create_plan(
        title="Build Dashboard",
        steps=[
            {"description": "Set up React project"},
            {"description": "Create layout components"},
            {"description": "Add Chart.js charts"},
            {"description": "Test and deploy"},
        ],
        goal="Build React dashboard with charts",
        conversation_id=conv_id,
    )

    # 5. Agent works on step 1
    await planner.update_step(plan["id"], 0, "in_progress")
    await conversation.add_assistant_message(content="Setting up the React project...")
    await planner.update_step(plan["id"], 0, "done", result="Created with Vite + React")

    # 6. Agent saves a checkpoint
    await tool.execute(
        action="checkpoint",
        checkpoint_data={
            "working_on": "Create layout components",
            "done": ["Set up React project"],
            "next": ["Add Chart.js charts", "Test and deploy"],
        },
    )

    # 7. Build messages — should include plan, facts, goals
    messages = await conversation.build_messages()

    # Verify plan is in context
    plan_in_context = any(
        "Active Plan" in (m.get("content") or "")
        for m in messages
        if m["role"] == "system"
    )
    assert plan_in_context, "Active plan should be in context"

    # Verify facts are in context
    facts_in_context = any(
        "Known facts" in (m.get("content") or "")
        for m in messages
        if m["role"] == "system"
    )
    assert facts_in_context, "Facts should be in context"

    # 8. Verify stats
    stats = await memory.get_memory_stats()
    assert stats["conversations"] == 1
    assert stats["facts"] == 2
    assert stats["active_goals"] == 1
    assert stats["checkpoints"] == 1

    # 9. Complete the goal
    await tool.execute(action="complete_goal", goal_id=goal_data["goal_id"])
    stats = await memory.get_memory_stats()
    assert stats["active_goals"] == 0



"""Persistent memory store backed by SQLite.

Stores conversation history, learned facts, conversation summaries,
goals, and memory checkpoints. This is the foundation of Plutus's
"never forget" system — all persistent state lives here.
"""

from __future__ import annotations

import json
import time
from typing import Any

import aiosqlite

_SCHEMA = """
CREATE TABLE IF NOT EXISTS conversations (
    id TEXT PRIMARY KEY,
    created_at REAL NOT NULL,
    title TEXT,
    metadata TEXT DEFAULT '{}',
    last_activity REAL
);

CREATE TABLE IF NOT EXISTS messages (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    conversation_id TEXT NOT NULL,
    role TEXT NOT NULL,
    content TEXT,
    tool_calls TEXT,
    tool_call_id TEXT,
    created_at REAL NOT NULL,
    FOREIGN KEY (conversation_id) REFERENCES conversations(id)
);

CREATE TABLE IF NOT EXISTS facts (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    category TEXT NOT NULL,
    content TEXT NOT NULL,
    source TEXT,
    created_at REAL NOT NULL,
    updated_at REAL NOT NULL
);

CREATE TABLE IF NOT EXISTS conversation_summaries (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    conversation_id TEXT NOT NULL UNIQUE,
    summary_data TEXT NOT NULL DEFAULT '{}',
    created_at REAL NOT NULL,
    updated_at REAL NOT NULL,
    FOREIGN KEY (conversation_id) REFERENCES conversations(id)
);

CREATE TABLE IF NOT EXISTS goals (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    conversation_id TEXT,
    description TEXT NOT NULL,
    status TEXT NOT NULL DEFAULT 'active',
    priority INTEGER NOT NULL DEFAULT 0,
    parent_goal_id INTEGER,
    created_at REAL NOT NULL,
    updated_at REAL NOT NULL,
    FOREIGN KEY (conversation_id) REFERENCES conversations(id),
    FOREIGN KEY (parent_goal_id) REFERENCES goals(id)
);

CREATE TABLE IF NOT EXISTS checkpoints (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    conversation_id TEXT NOT NULL,
    checkpoint_type TEXT NOT NULL DEFAULT 'auto',
    state_data TEXT NOT NULL DEFAULT '{}',
    message_id INTEGER,
    created_at REAL NOT NULL,
    FOREIGN KEY (conversation_id) REFERENCES conversations(id)
);

CREATE INDEX IF NOT EXISTS idx_messages_conversation ON messages(conversation_id);
CREATE INDEX IF NOT EXISTS idx_facts_category ON facts(category);
CREATE INDEX IF NOT EXISTS idx_summaries_conversation ON conversation_summaries(conversation_id);
CREATE INDEX IF NOT EXISTS idx_goals_conversation ON goals(conversation_id);
CREATE INDEX IF NOT EXISTS idx_goals_status ON goals(status);
CREATE INDEX IF NOT EXISTS idx_checkpoints_conversation ON checkpoints(conversation_id);
"""


class MemoryStore:
    """Async SQLite-backed memory store with summarization and goal support."""

    def __init__(self, db_path: str):
        self._db_path = db_path
        self._db: aiosqlite.Connection | None = None

    async def initialize(self) -> None:
        self._db = await aiosqlite.connect(self._db_path)
        await self._db.executescript(_SCHEMA)
        await self._db.commit()
        # Migrate: add last_activity column if missing (for existing databases)
        await self._migrate_last_activity()
        # Create index after migration ensures column exists
        await self._db.execute(
            "CREATE INDEX IF NOT EXISTS idx_conversations_last_activity ON conversations(last_activity)"
        )
        await self._db.commit()

    async def _migrate_last_activity(self) -> None:
        """Add last_activity column to conversations table if it doesn't exist."""
        assert self._db
        try:
            cursor = await self._db.execute("PRAGMA table_info(conversations)")
            columns = [row[1] for row in await cursor.fetchall()]
            if "last_activity" not in columns:
                await self._db.execute(
                    "ALTER TABLE conversations ADD COLUMN last_activity REAL"
                )
                # Backfill: set last_activity to the latest message time or created_at
                await self._db.execute(
                    "UPDATE conversations SET last_activity = "
                    "COALESCE((SELECT MAX(created_at) FROM messages WHERE conversation_id = conversations.id), created_at)"
                )
                await self._db.commit()
        except Exception:
            pass  # Column already exists or other non-critical error

    async def close(self) -> None:
        if self._db:
            await self._db.close()

    # -- Conversations --

    async def create_conversation(self, conv_id: str, title: str | None = None) -> None:
        assert self._db
        now = time.time()
        await self._db.execute(
            "INSERT INTO conversations (id, created_at, title, last_activity) VALUES (?, ?, ?, ?)",
            (conv_id, now, title, now),
        )
        await self._db.commit()

    async def list_conversations(self, limit: int = 50) -> list[dict[str, Any]]:
        assert self._db
        cursor = await self._db.execute(
            "SELECT c.id, c.created_at, c.title, c.metadata, c.last_activity, "
            "(SELECT COUNT(*) FROM messages m WHERE m.conversation_id = c.id) as msg_count "
            "FROM conversations c "
            "ORDER BY COALESCE(c.last_activity, c.created_at) DESC LIMIT ?",
            (limit,),
        )
        rows = await cursor.fetchall()
        return [
            {
                "id": r[0],
                "created_at": r[1],
                "title": r[2],
                "metadata": json.loads(r[3]) if r[3] else {},
                "last_activity": r[4] or r[1],
                "message_count": r[5],
            }
            for r in rows
        ]

    async def rename_conversation(self, conv_id: str, title: str) -> None:
        """Rename a conversation."""
        assert self._db
        await self._db.execute(
            "UPDATE conversations SET title = ? WHERE id = ?",
            (title, conv_id),
        )
        await self._db.commit()

    async def touch_conversation(self, conv_id: str) -> None:
        """Update last_activity timestamp for a conversation."""
        assert self._db
        await self._db.execute(
            "UPDATE conversations SET last_activity = ? WHERE id = ?",
            (time.time(), conv_id),
        )
        await self._db.commit()

    async def delete_conversation(self, conv_id: str) -> None:
        assert self._db
        await self._db.execute("DELETE FROM messages WHERE conversation_id = ?", (conv_id,))
        await self._db.execute("DELETE FROM conversation_summaries WHERE conversation_id = ?", (conv_id,))
        await self._db.execute("DELETE FROM goals WHERE conversation_id = ?", (conv_id,))
        await self._db.execute("DELETE FROM checkpoints WHERE conversation_id = ?", (conv_id,))
        await self._db.execute("DELETE FROM conversations WHERE id = ?", (conv_id,))
        await self._db.commit()

    async def cleanup_stale_conversations(self, max_age_days: int = 30) -> int:
        """Delete conversations with no activity for more than max_age_days.

        Returns the number of conversations deleted.
        """
        assert self._db
        cutoff = time.time() - (max_age_days * 86400)

        # Find stale conversations
        cursor = await self._db.execute(
            "SELECT id FROM conversations WHERE COALESCE(last_activity, created_at) < ?",
            (cutoff,),
        )
        stale_ids = [r[0] for r in await cursor.fetchall()]

        if not stale_ids:
            return 0

        # Delete all related data for each stale conversation
        placeholders = ",".join("?" for _ in stale_ids)
        await self._db.execute(
            f"DELETE FROM messages WHERE conversation_id IN ({placeholders})",
            stale_ids,
        )
        await self._db.execute(
            f"DELETE FROM conversation_summaries WHERE conversation_id IN ({placeholders})",
            stale_ids,
        )
        await self._db.execute(
            f"DELETE FROM goals WHERE conversation_id IN ({placeholders})",
            stale_ids,
        )
        await self._db.execute(
            f"DELETE FROM checkpoints WHERE conversation_id IN ({placeholders})",
            stale_ids,
        )
        await self._db.execute(
            f"DELETE FROM conversations WHERE id IN ({placeholders})",
            stale_ids,
        )
        await self._db.commit()
        return len(stale_ids)

    # -- Messages --

    async def add_message(
        self,
        conversation_id: str,
        role: str,
        content: str | None = None,
        tool_calls: list[dict] | None = None,
        tool_call_id: str | None = None,
    ) -> int:
        assert self._db
        now = time.time()
        cursor = await self._db.execute(
            "INSERT INTO messages (conversation_id, role, content, tool_calls, tool_call_id, created_at) "
            "VALUES (?, ?, ?, ?, ?, ?)",
            (
                conversation_id,
                role,
                content,
                json.dumps(tool_calls) if tool_calls else None,
                tool_call_id,
                now,
            ),
        )
        # Update last_activity on the conversation
        await self._db.execute(
            "UPDATE conversations SET last_activity = ? WHERE id = ?",
            (now, conversation_id),
        )
        await self._db.commit()
        return cursor.lastrowid  # type: ignore

    async def get_messages(
        self, conversation_id: str, limit: int | None = None
    ) -> list[dict[str, Any]]:
        assert self._db
        query = (
            "SELECT id, role, content, tool_calls, tool_call_id, created_at "
            "FROM messages WHERE conversation_id = ? ORDER BY created_at ASC"
        )
        params: tuple = (conversation_id,)
        if limit:
            query = (
                "SELECT * FROM ("
                "  SELECT id, role, content, tool_calls, tool_call_id, created_at "
                "  FROM messages WHERE conversation_id = ? "
                "  ORDER BY created_at DESC LIMIT ?"
                ") sub ORDER BY created_at ASC"
            )
            params = (conversation_id, limit)

        cursor = await self._db.execute(query, params)
        rows = await cursor.fetchall()
        return [
            {
                "id": r[0],
                "role": r[1],
                "content": r[2],
                "tool_calls": json.loads(r[3]) if r[3] else None,
                "tool_call_id": r[4],
                "created_at": r[5],
            }
            for r in rows
        ]

    async def get_message_count(self, conversation_id: str) -> int:
        """Get the total number of messages in a conversation."""
        assert self._db
        cursor = await self._db.execute(
            "SELECT COUNT(*) FROM messages WHERE conversation_id = ?",
            (conversation_id,),
        )
        row = await cursor.fetchone()
        return row[0] if row else 0

    # -- Facts (persistent memory) --

    async def store_fact(self, category: str, content: str, source: str | None = None) -> int:
        assert self._db
        now = time.time()

        # Check for duplicate facts (same category + content)
        cursor = await self._db.execute(
            "SELECT id FROM facts WHERE category = ? AND content = ?",
            (category, content),
        )
        existing = await cursor.fetchone()
        if existing:
            # Update the timestamp
            await self._db.execute(
                "UPDATE facts SET updated_at = ? WHERE id = ?",
                (now, existing[0]),
            )
            await self._db.commit()
            return existing[0]

        cursor = await self._db.execute(
            "INSERT INTO facts (category, content, source, created_at, updated_at) "
            "VALUES (?, ?, ?, ?, ?)",
            (category, content, source, now, now),
        )
        await self._db.commit()
        return cursor.lastrowid  # type: ignore

    async def get_facts(self, category: str | None = None, limit: int = 50) -> list[dict[str, Any]]:
        assert self._db
        if category:
            cursor = await self._db.execute(
                "SELECT id, category, content, source, created_at FROM facts "
                "WHERE category = ? ORDER BY updated_at DESC LIMIT ?",
                (category, limit),
            )
        else:
            cursor = await self._db.execute(
                "SELECT id, category, content, source, created_at FROM facts "
                "ORDER BY updated_at DESC LIMIT ?",
                (limit,),
            )
        rows = await cursor.fetchall()
        return [
            {
                "id": r[0],
                "category": r[1],
                "content": r[2],
                "source": r[3],
                "created_at": r[4],
            }
            for r in rows
        ]

    async def search_facts(self, query: str, limit: int = 10) -> list[dict[str, Any]]:
        """Search facts by content (simple LIKE query)."""
        assert self._db
        cursor = await self._db.execute(
            "SELECT id, category, content, source, created_at FROM facts "
            "WHERE content LIKE ? ORDER BY updated_at DESC LIMIT ?",
            (f"%{query}%", limit),
        )
        rows = await cursor.fetchall()
        return [
            {
                "id": r[0],
                "category": r[1],
                "content": r[2],
                "source": r[3],
                "created_at": r[4],
            }
            for r in rows
        ]

    async def delete_fact(self, fact_id: int) -> None:
        assert self._db
        await self._db.execute("DELETE FROM facts WHERE id = ?", (fact_id,))
        await self._db.commit()

    # -- Conversation Summaries --

    async def save_conversation_summary(
        self, conversation_id: str, summary_data: dict[str, Any]
    ) -> None:
        """Save or update a conversation summary."""
        assert self._db
        now = time.time()
        summary_json = json.dumps(summary_data)

        # Upsert
        await self._db.execute(
            """INSERT INTO conversation_summaries (conversation_id, summary_data, created_at, updated_at)
               VALUES (?, ?, ?, ?)
               ON CONFLICT(conversation_id) DO UPDATE SET
                 summary_data = excluded.summary_data,
                 updated_at = excluded.updated_at""",
            (conversation_id, summary_json, now, now),
        )
        await self._db.commit()

    async def get_conversation_summary(
        self, conversation_id: str
    ) -> dict[str, Any] | None:
        """Get the summary for a conversation."""
        assert self._db
        cursor = await self._db.execute(
            "SELECT summary_data FROM conversation_summaries WHERE conversation_id = ?",
            (conversation_id,),
        )
        row = await cursor.fetchone()
        if not row:
            return None
        return json.loads(row[0])

    # -- Goals --

    async def add_goal(
        self,
        description: str,
        conversation_id: str | None = None,
        priority: int = 0,
        parent_goal_id: int | None = None,
    ) -> int:
        """Add a new goal."""
        assert self._db
        now = time.time()
        cursor = await self._db.execute(
            "INSERT INTO goals (conversation_id, description, status, priority, parent_goal_id, created_at, updated_at) "
            "VALUES (?, ?, 'active', ?, ?, ?, ?)",
            (conversation_id, description, priority, parent_goal_id, now, now),
        )
        await self._db.commit()
        return cursor.lastrowid  # type: ignore

    async def get_active_goals(
        self, conversation_id: str | None = None, limit: int = 20
    ) -> list[dict[str, Any]]:
        """Get active goals, optionally scoped to a conversation."""
        assert self._db
        if conversation_id:
            cursor = await self._db.execute(
                "SELECT id, conversation_id, description, status, priority, parent_goal_id, created_at "
                "FROM goals WHERE status = 'active' AND conversation_id = ? "
                "ORDER BY priority DESC, created_at ASC LIMIT ?",
                (conversation_id, limit),
            )
        else:
            cursor = await self._db.execute(
                "SELECT id, conversation_id, description, status, priority, parent_goal_id, created_at "
                "FROM goals WHERE status = 'active' "
                "ORDER BY priority DESC, created_at ASC LIMIT ?",
                (limit,),
            )
        rows = await cursor.fetchall()
        return [
            {
                "id": r[0],
                "conversation_id": r[1],
                "description": r[2],
                "status": r[3],
                "priority": r[4],
                "parent_goal_id": r[5],
                "created_at": r[6],
            }
            for r in rows
        ]

    async def update_goal_status(self, goal_id: int, status: str) -> None:
        """Update a goal's status (active, completed, failed, cancelled)."""
        assert self._db
        await self._db.execute(
            "UPDATE goals SET status = ?, updated_at = ? WHERE id = ?",
            (status, time.time(), goal_id),
        )
        await self._db.commit()

    async def get_all_goals(
        self, conversation_id: str | None = None, limit: int = 50
    ) -> list[dict[str, Any]]:
        """Get all goals regardless of status."""
        assert self._db
        if conversation_id:
            cursor = await self._db.execute(
                "SELECT id, conversation_id, description, status, priority, parent_goal_id, created_at "
                "FROM goals WHERE conversation_id = ? ORDER BY created_at DESC LIMIT ?",
                (conversation_id, limit),
            )
        else:
            cursor = await self._db.execute(
                "SELECT id, conversation_id, description, status, priority, parent_goal_id, created_at "
                "FROM goals ORDER BY created_at DESC LIMIT ?",
                (limit,),
            )
        rows = await cursor.fetchall()
        return [
            {
                "id": r[0],
                "conversation_id": r[1],
                "description": r[2],
                "status": r[3],
                "priority": r[4],
                "parent_goal_id": r[5],
                "created_at": r[6],
            }
            for r in rows
        ]

    # -- Checkpoints --

    async def save_checkpoint(
        self,
        conversation_id: str,
        state_data: dict[str, Any],
        checkpoint_type: str = "auto",
        message_id: int | None = None,
    ) -> int:
        """Save a checkpoint of the current state."""
        assert self._db
        cursor = await self._db.execute(
            "INSERT INTO checkpoints (conversation_id, checkpoint_type, state_data, message_id, created_at) "
            "VALUES (?, ?, ?, ?, ?)",
            (conversation_id, checkpoint_type, json.dumps(state_data), message_id, time.time()),
        )
        await self._db.commit()
        return cursor.lastrowid  # type: ignore

    async def get_latest_checkpoint(
        self, conversation_id: str
    ) -> dict[str, Any] | None:
        """Get the most recent checkpoint for a conversation."""
        assert self._db
        cursor = await self._db.execute(
            "SELECT id, checkpoint_type, state_data, message_id, created_at "
            "FROM checkpoints WHERE conversation_id = ? ORDER BY created_at DESC LIMIT 1",
            (conversation_id,),
        )
        row = await cursor.fetchone()
        if not row:
            return None
        return {
            "id": row[0],
            "checkpoint_type": row[1],
            "state_data": json.loads(row[2]),
            "message_id": row[3],
            "created_at": row[4],
        }

    async def list_checkpoints(
        self, conversation_id: str, limit: int = 10
    ) -> list[dict[str, Any]]:
        """List checkpoints for a conversation."""
        assert self._db
        cursor = await self._db.execute(
            "SELECT id, checkpoint_type, state_data, message_id, created_at "
            "FROM checkpoints WHERE conversation_id = ? ORDER BY created_at DESC LIMIT ?",
            (conversation_id, limit),
        )
        rows = await cursor.fetchall()
        return [
            {
                "id": r[0],
                "checkpoint_type": r[1],
                "state_data": json.loads(r[2]),
                "message_id": r[3],
                "created_at": r[4],
            }
            for r in rows
        ]

    # -- Memory stats --

    async def get_memory_stats(self) -> dict[str, Any]:
        """Get overall memory statistics."""
        assert self._db

        conv_count = await self._db.execute("SELECT COUNT(*) FROM conversations")
        msg_count = await self._db.execute("SELECT COUNT(*) FROM messages")
        fact_count = await self._db.execute("SELECT COUNT(*) FROM facts")
        goal_count = await self._db.execute("SELECT COUNT(*) FROM goals WHERE status = 'active'")
        summary_count = await self._db.execute("SELECT COUNT(*) FROM conversation_summaries")
        checkpoint_count = await self._db.execute("SELECT COUNT(*) FROM checkpoints")

        return {
            "conversations": (await conv_count.fetchone())[0],
            "messages": (await msg_count.fetchone())[0],
            "facts": (await fact_count.fetchone())[0],
            "active_goals": (await goal_count.fetchone())[0],
            "summaries": (await summary_count.fetchone())[0],
            "checkpoints": (await checkpoint_count.fetchone())[0],
        }



"""Context Awareness Engine — always knows where Plutus is.

This module solves the #1 problem with computer-use agents: acting on the wrong
window. Before every action, Plutus checks:
  1. Which window is currently active (title, app, PID)
  2. Whether that matches the intended target
  3. If not, it auto-focuses the correct window first

The context is also injected into every tool result so the LLM always knows
exactly what app/window it's looking at.

Architecture:
  ContextState  — snapshot of the current desktop state
  ContextEngine — singleton that tracks state and provides guards
  ActionGuard   — wraps every pc action with pre/post context checks
"""

from __future__ import annotations

import asyncio
import json
import logging
import platform
import re
import time
from dataclasses import dataclass, field
from typing import Any

SYSTEM = platform.system()
logger = logging.getLogger("plutus.pc.context")


# ─────────────────────────────────────────────────────────────
# Context State — a snapshot of what's happening on the desktop
# ─────────────────────────────────────────────────────────────

@dataclass
class ContextState:
    """Snapshot of the current desktop context."""

    # Active window info
    active_window_title: str = ""
    active_app_name: str = ""
    active_window_pid: int = 0

    # Derived context
    active_app_category: str = ""  # browser, editor, terminal, messenger, etc.
    active_browser_tab: str = ""   # if in a browser, what tab/URL
    active_document: str = ""      # if in an editor, what file

    # Screen state
    screen_width: int = 0
    screen_height: int = 0
    mouse_x: int = 0
    mouse_y: int = 0

    # Timestamps
    captured_at: float = 0.0
    age_seconds: float = 0.0

    # Confidence
    is_stale: bool = False  # True if context is older than threshold

    def to_dict(self) -> dict[str, Any]:
        return {
            "active_window": {
                "title": self.active_window_title,
                "app": self.active_app_name,
                "pid": self.active_window_pid,
                "category": self.active_app_category,
            },
            "browser_tab": self.active_browser_tab or None,
            "active_document": self.active_document or None,
            "screen": {
                "width": self.screen_width,
                "height": self.screen_height,
            },
            "mouse": {"x": self.mouse_x, "y": self.mouse_y},
            "captured_at": self.captured_at,
            "is_stale": self.is_stale,
        }

    def summary(self) -> str:
        """Human-readable one-line summary for injecting into LLM context."""
        parts = []
        if self.active_app_name:
            parts.append(f"Active app: {self.active_app_name}")
        if self.active_window_title:
            title = self.active_window_title
            if len(title) > 80:
                title = title[:77] + "..."
            parts.append(f"Window: \"{title}\"")
        if self.active_app_category:
            parts.append(f"Category: {self.active_app_category}")
        if self.active_browser_tab:
            parts.append(f"Tab: {self.active_browser_tab}")
        if self.active_document:
            parts.append(f"Document: {self.active_document}")
        return " | ".join(parts) if parts else "No active window detected"


# ─────────────────────────────────────────────────────────────
# App Category Detection — classify what kind of app is active
# ─────────────────────────────────────────────────────────────

# Map of known apps to categories
APP_CATEGORIES: dict[str, str] = {
    # Browsers
    "chrome": "browser", "firefox": "browser", "edge": "browser",
    "msedge": "browser", "safari": "browser", "brave": "browser",
    "opera": "browser", "vivaldi": "browser", "arc": "browser",
    # Messengers
    "whatsapp": "messenger", "telegram": "messenger", "signal": "messenger",
    "discord": "messenger", "slack": "messenger", "teams": "messenger",
    "skype": "messenger", "messenger": "messenger", "wechat": "messenger",
    "viber": "messenger", "line": "messenger", "imessage": "messenger",
    # Editors / IDEs
    "code": "editor", "vscode": "editor", "visual studio code": "editor",
    "sublime": "editor", "atom": "editor", "notepad": "editor",
    "notepad++": "editor", "vim": "editor", "neovim": "editor",
    "emacs": "editor", "pycharm": "editor", "intellij": "editor",
    "webstorm": "editor", "cursor": "editor",
    # Terminals
    "terminal": "terminal", "iterm": "terminal", "iterm2": "terminal",
    "cmd": "terminal", "powershell": "terminal", "windowsterminal": "terminal",
    "windows terminal": "terminal", "wt": "terminal", "alacritty": "terminal",
    "kitty": "terminal", "hyper": "terminal", "warp": "terminal",
    "gnome-terminal": "terminal", "konsole": "terminal",
    # Office
    "word": "office", "winword": "office", "excel": "office",
    "powerpoint": "office", "powerpnt": "office", "onenote": "office",
    "outlook": "office", "libreoffice": "office", "pages": "office",
    "numbers": "office", "keynote": "office", "google docs": "office",
    # File managers
    "explorer": "file_manager", "finder": "file_manager",
    "nautilus": "file_manager", "dolphin": "file_manager",
    "thunar": "file_manager", "nemo": "file_manager",
    # Media
    "spotify": "media", "vlc": "media", "itunes": "media",
    "music": "media", "photos": "media", "preview": "media",
    # AI / Chat
    "chatgpt": "ai_chat", "claude": "ai_chat", "copilot": "ai_chat",
    "bard": "ai_chat", "perplexity": "ai_chat",
    # System
    "settings": "system", "systempreferences": "system",
    "control panel": "system", "task manager": "system",
    "activity monitor": "system",
}

# Patterns for detecting browser tab content from window titles
BROWSER_TITLE_PATTERNS = [
    # "Page Title - Google Chrome"
    re.compile(r'^(.+?)\s*[-–—]\s*(?:Google Chrome|Mozilla Firefox|Microsoft Edge|Safari|Brave|Opera|Vivaldi|Arc)$', re.I),
    # "Page Title — Browser"
    re.compile(r'^(.+?)\s*[-–—]\s*\w+\s*Browser$', re.I),
]

# Patterns for detecting document names from editor titles
EDITOR_TITLE_PATTERNS = [
    # "file.py - Visual Studio Code"
    re.compile(r'^(.+?\.[\w]+)\s*[-–—]', re.I),
    # "file.py — Sublime Text"
    re.compile(r'^(.+?\.[\w]+)\s', re.I),
]

# Patterns for detecting messenger context from window titles
MESSENGER_TITLE_PATTERNS = [
    # "Contact Name - WhatsApp"
    re.compile(r'^(.+?)\s*[-–—]\s*(?:WhatsApp|Telegram|Signal|Discord|Slack|Teams)', re.I),
    # "Chat with Contact"
    re.compile(r'^(?:Chat with|Conversation with)\s+(.+)', re.I),
]


def classify_app(app_name: str, window_title: str = "") -> dict[str, str]:
    """Classify an app and extract context from its window title.

    Returns:
        {
            "category": "browser" | "messenger" | "editor" | etc.,
            "browser_tab": "..." (if browser),
            "document": "..." (if editor),
            "chat_context": "..." (if messenger),
        }
    """
    result: dict[str, str] = {"category": "", "browser_tab": "", "document": "", "chat_context": ""}

    app_lower = app_name.lower().strip()
    title_lower = window_title.lower().strip()

    # Direct app name match
    for key, category in APP_CATEGORIES.items():
        if key in app_lower:
            result["category"] = category
            break

    # If not found by app name, try window title
    if not result["category"]:
        for key, category in APP_CATEGORIES.items():
            if key in title_lower:
                result["category"] = category
                break

    # Extract browser tab from title
    if result["category"] == "browser":
        for pattern in BROWSER_TITLE_PATTERNS:
            match = pattern.match(window_title)
            if match:
                result["browser_tab"] = match.group(1).strip()
                break
        # Check if browser is showing a known web app
        if any(x in title_lower for x in ["chatgpt", "claude.ai", "bard.google"]):
            result["category"] = "ai_chat"
            result["browser_tab"] = window_title.split(" - ")[0].strip() if " - " in window_title else window_title
        elif any(x in title_lower for x in ["web.whatsapp", "web.telegram", "discord.com/channels"]):
            result["category"] = "messenger"
            result["browser_tab"] = window_title.split(" - ")[0].strip() if " - " in window_title else window_title

    # Extract document from editor title
    elif result["category"] == "editor":
        for pattern in EDITOR_TITLE_PATTERNS:
            match = pattern.match(window_title)
            if match:
                result["document"] = match.group(1).strip()
                break

    # Extract chat context from messenger title
    elif result["category"] == "messenger":
        for pattern in MESSENGER_TITLE_PATTERNS:
            match = pattern.match(window_title)
            if match:
                result["chat_context"] = match.group(1).strip()
                break

    return result


# ─────────────────────────────────────────────────────────────
# Context Engine — the singleton that tracks desktop state
# ─────────────────────────────────────────────────────────────

class ContextEngine:
    """Tracks the current desktop context and provides pre-action guards.

    Usage:
        ctx = ContextEngine()
        state = await ctx.get_context()       # get current state
        await ctx.ensure_app("WhatsApp")      # focus WhatsApp if not active
        result = await ctx.guarded_action(     # run action with context check
            target_app="WhatsApp",
            action_fn=some_async_fn,
        )
    """

    # How old context can be before it's considered stale
    STALE_THRESHOLD = 3.0  # seconds

    # How long to wait after focusing a window before acting
    FOCUS_SETTLE_TIME = 0.5  # seconds

    def __init__(self):
        self._last_state: ContextState | None = None
        self._state_lock = asyncio.Lock()
        self._focus_history: list[dict[str, Any]] = []  # track recent focus changes
        self._action_log: list[dict[str, Any]] = []     # track recent actions + their context

    async def _run_cmd(self, cmd: str, timeout: float = 5) -> str:
        """Run a shell command and return stdout."""
        try:
            proc = await asyncio.create_subprocess_shell(
                cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, _ = await asyncio.wait_for(proc.communicate(), timeout=timeout)
            return stdout.decode("utf-8", errors="replace").strip()
        except Exception:
            return ""

    # ─── Get Active Window (cross-platform) ───

    async def _get_active_window_windows(self) -> dict[str, Any]:
        """Get the active window on Windows using PowerShell."""
        cmd = (
            'powershell -NoProfile -Command "'
            '$fw = [System.Diagnostics.Process]::GetProcesses() | '
            'Where-Object { $_.MainWindowHandle -eq '
            '(Add-Type -MemberDefinition \'[DllImport(\"user32.dll\")] '
            'public static extern IntPtr GetForegroundWindow();\' '
            '-Name Win32 -PassThru)::GetForegroundWindow() }; '
            'if ($fw) { @{Title=$fw.MainWindowTitle; App=$fw.ProcessName; PID=$fw.Id} | ConvertTo-Json } '
            'else { \'{\"Title\":\"\",\"App\":\"\",\"PID\":0}\' }"'
        )
        out = await self._run_cmd(cmd)
        try:
            data = json.loads(out)
            return {
                "title": data.get("Title", ""),
                "app": data.get("App", ""),
                "pid": data.get("PID", 0),
            }
        except Exception:
            return {"title": "", "app": "", "pid": 0}

    async def _get_active_window_macos(self) -> dict[str, Any]:
        """Get the active window on macOS using AppleScript."""
        script = '''osascript -e '
        tell application "System Events"
            set frontApp to first application process whose frontmost is true
            set appName to name of frontApp
            try
                set winTitle to name of front window of frontApp
            on error
                set winTitle to ""
            end try
            set appPID to unix id of frontApp
        end tell
        return appName & "|" & winTitle & "|" & appPID'
        '''
        out = await self._run_cmd(script)
        parts = out.split("|")
        if len(parts) >= 3:
            return {
                "title": parts[1].strip(),
                "app": parts[0].strip(),
                "pid": int(parts[2].strip()) if parts[2].strip().isdigit() else 0,
            }
        return {"title": "", "app": "", "pid": 0}

    async def _get_active_window_linux(self) -> dict[str, Any]:
        """Get the active window on Linux using xdotool/xprop."""
        # Try xdotool first
        wid = await self._run_cmd("xdotool getactivewindow 2>/dev/null")
        if wid:
            title = await self._run_cmd(f"xdotool getwindowname {wid} 2>/dev/null")
            pid = await self._run_cmd(f"xdotool getwindowpid {wid} 2>/dev/null")
            # Try to get app name from PID
            app = ""
            if pid and pid.isdigit():
                app = await self._run_cmd(f"ps -p {pid} -o comm= 2>/dev/null")
            return {
                "title": title,
                "app": app,
                "pid": int(pid) if pid and pid.isdigit() else 0,
            }

        # Fallback: xprop
        out = await self._run_cmd(
            "xprop -root _NET_ACTIVE_WINDOW 2>/dev/null | grep -oP '0x[0-9a-f]+'"
        )
        if out:
            title = await self._run_cmd(f"xprop -id {out} WM_NAME 2>/dev/null")
            match = re.search(r'"(.+)"', title)
            return {
                "title": match.group(1) if match else "",
                "app": "",
                "pid": 0,
            }

        return {"title": "", "app": "", "pid": 0}

    async def get_active_window(self) -> dict[str, Any]:
        """Get the currently active/foreground window (cross-platform)."""
        if SYSTEM == "Windows":
            return await self._get_active_window_windows()
        elif SYSTEM == "Darwin":
            return await self._get_active_window_macos()
        else:
            return await self._get_active_window_linux()

    # ─── Get Full Context ───

    async def get_context(self, force_refresh: bool = False) -> ContextState:
        """Get the current desktop context.

        Returns a cached state if it's fresh enough, otherwise refreshes.
        """
        async with self._state_lock:
            now = time.time()

            # Return cached if fresh
            if (
                not force_refresh
                and self._last_state
                and (now - self._last_state.captured_at) < self.STALE_THRESHOLD
            ):
                self._last_state.age_seconds = now - self._last_state.captured_at
                return self._last_state

            # Refresh
            window = await self.get_active_window()
            classification = classify_app(
                window.get("app", ""),
                window.get("title", ""),
            )

            # Get mouse position
            mouse_x, mouse_y = 0, 0
            try:
                import pyautogui
                pos = pyautogui.position()
                mouse_x, mouse_y = pos.x, pos.y
                size = pyautogui.size()
                screen_w, screen_h = size.width, size.height
            except Exception:
                screen_w, screen_h = 1920, 1080

            state = ContextState(
                active_window_title=window.get("title", ""),
                active_app_name=window.get("app", ""),
                active_window_pid=window.get("pid", 0),
                active_app_category=classification["category"],
                active_browser_tab=classification["browser_tab"],
                active_document=classification["document"],
                screen_width=screen_w,
                screen_height=screen_h,
                mouse_x=mouse_x,
                mouse_y=mouse_y,
                captured_at=now,
                age_seconds=0.0,
                is_stale=False,
            )

            self._last_state = state
            return state

    # ─── Focus Verification & Auto-Focus ───

    async def is_app_active(self, target: str) -> bool:
        """Check if a specific app/window is currently in the foreground.

        Matches against app name, window title, and category.
        Uses fuzzy matching for convenience.
        """
        state = await self.get_context(force_refresh=True)
        target_lower = target.lower().strip()

        # Check app name
        if target_lower in state.active_app_name.lower():
            return True

        # Check window title
        if target_lower in state.active_window_title.lower():
            return True

        # Check category
        if target_lower == state.active_app_category:
            return True

        # Check browser tab (for web apps)
        if state.active_browser_tab and target_lower in state.active_browser_tab.lower():
            return True

        return False

    async def ensure_app(self, target: str, timeout: float = 5.0) -> dict[str, Any]:
        """Ensure a specific app is in the foreground. Focus it if not.

        This is the KEY method that prevents wrong-window actions.
        Call this before any action that targets a specific app.

        Returns:
            {
                "already_active": bool,
                "focused": bool,
                "app": str,
                "title": str,
                "error": str | None,
            }
        """
        # Check if already active
        if await self.is_app_active(target):
            state = await self.get_context()
            return {
                "already_active": True,
                "focused": True,
                "app": state.active_app_name,
                "title": state.active_window_title,
                "error": None,
            }

        # Need to focus — import WindowManager
        from plutus.pc.windows import WindowManager
        wm = WindowManager()

        result = await wm.focus(target)

        if result.get("success"):
            # Wait for the window to settle
            await asyncio.sleep(self.FOCUS_SETTLE_TIME)

            # Verify focus actually changed
            state = await self.get_context(force_refresh=True)

            # Log the focus change
            self._focus_history.append({
                "target": target,
                "result_app": state.active_app_name,
                "result_title": state.active_window_title,
                "timestamp": time.time(),
                "success": True,
            })

            logger.info(
                f"Focused window: {state.active_app_name} - {state.active_window_title}"
            )

            return {
                "already_active": False,
                "focused": True,
                "app": state.active_app_name,
                "title": state.active_window_title,
                "error": None,
            }
        else:
            self._focus_history.append({
                "target": target,
                "result_app": "",
                "result_title": "",
                "timestamp": time.time(),
                "success": False,
                "error": result.get("error", "Unknown error"),
            })

            return {
                "already_active": False,
                "focused": False,
                "app": "",
                "title": "",
                "error": result.get("error", f"Could not focus: {target}"),
            }

    # ─── Context-Enriched Results ───

    def enrich_result(self, result: dict[str, Any], operation: str) -> dict[str, Any]:
        """Add current context info to a tool result.

        This is injected into EVERY pc tool result so the LLM always knows
        what window/app it's looking at.
        """
        if self._last_state:
            result["_context"] = {
                "active_app": self._last_state.active_app_name,
                "active_window": self._last_state.active_window_title[:100],
                "category": self._last_state.active_app_category,
                "mouse_at": f"({self._last_state.mouse_x}, {self._last_state.mouse_y})",
            }
            if self._last_state.active_browser_tab:
                result["_context"]["browser_tab"] = self._last_state.active_browser_tab
            if self._last_state.active_document:
                result["_context"]["document"] = self._last_state.active_document
        return result

    # ─── Action Logging ───

    def log_action(self, operation: str, params: dict[str, Any], result: Any) -> None:
        """Log an action with its context for debugging and learning."""
        entry = {
            "operation": operation,
            "params": {k: v for k, v in params.items() if k != "include_base64"},
            "context": self._last_state.summary() if self._last_state else "unknown",
            "timestamp": time.time(),
        }
        self._action_log.append(entry)

        # Keep only last 100 actions
        if len(self._action_log) > 100:
            self._action_log = self._action_log[-100:]

    # ─── Status / Debug ───

    def get_status(self) -> dict[str, Any]:
        """Get the context engine status for the UI/API."""
        return {
            "last_state": self._last_state.to_dict() if self._last_state else None,
            "focus_history": self._focus_history[-10:],
            "recent_actions": self._action_log[-10:],
            "stale_threshold": self.STALE_THRESHOLD,
            "focus_settle_time": self.FOCUS_SETTLE_TIME,
        }


# ─────────────────────────────────────────────────────────────
# Action Guard — wraps pc actions with context verification
# ─────────────────────────────────────────────────────────────

class ActionGuard:
    """Pre-action guard that ensures context is correct before acting.

    Wraps around the PCControlTool to intercept actions and:
    1. Refresh context before the action
    2. If a target app is specified, ensure it's focused
    3. Enrich the result with current context
    4. Log the action

    Usage:
        guard = ActionGuard(context_engine)
        result = await guard.check_before_action(
            operation="type",
            params={"text": "hello"},
            target_app="WhatsApp",  # optional
        )
    """

    # Operations that modify state (need context verification)
    WRITE_OPERATIONS = {
        "click", "double_click", "right_click", "drag",
        "type", "press", "hotkey", "shortcut", "key_down",
        "scroll",
    }

    # Operations that only read state (less critical)
    READ_OPERATIONS = {
        "screenshot", "read_screen", "find_text", "find_elements",
        "get_pixel_color", "find_color", "wait_for_text", "wait_for_change",
        "screen_info", "list_windows", "find_window", "active_window",
        "list_shortcuts", "list_workflows", "list_templates", "get_template",
    }

    def __init__(self, context_engine: ContextEngine):
        self._ctx = context_engine

    async def check_before_action(
        self,
        operation: str,
        params: dict[str, Any],
        target_app: str | None = None,
    ) -> dict[str, Any]:
        """Run pre-action checks and return context info.

        Returns:
            {
                "proceed": bool,         # whether to proceed with the action
                "context": ContextState,  # current context
                "focus_result": dict,     # if focus was needed
                "warning": str | None,    # any warnings
            }
        """
        # Always refresh context for write operations
        force_refresh = operation in self.WRITE_OPERATIONS
        context = await self._ctx.get_context(force_refresh=force_refresh)

        result: dict[str, Any] = {
            "proceed": True,
            "context": context,
            "focus_result": None,
            "warning": None,
        }

        # If a target app is specified, ensure it's focused
        if target_app and operation in self.WRITE_OPERATIONS:
            if not await self._ctx.is_app_active(target_app):
                focus_result = await self._ctx.ensure_app(target_app)
                result["focus_result"] = focus_result

                if not focus_result.get("focused"):
                    result["proceed"] = False
                    result["warning"] = (
                        f"Cannot focus {target_app}: {focus_result.get('error')}. "
                        f"Currently active: {context.active_app_name} - {context.active_window_title}"
                    )
                    return result

                # Re-get context after focus change
                result["context"] = await self._ctx.get_context(force_refresh=True)

        return result

    async def post_action(
        self,
        operation: str,
        params: dict[str, Any],
        result: dict[str, Any],
    ) -> dict[str, Any]:
        """Enrich the result with context after an action."""
        # Refresh context after write operations
        if operation in self.WRITE_OPERATIONS:
            await self._ctx.get_context(force_refresh=True)

        # Enrich result with context
        enriched = self._ctx.enrich_result(result, operation)

        # Log the action
        self._ctx.log_action(operation, params, result)

        return enriched


# ─────────────────────────────────────────────────────────────
# Singleton — global context engine instance
# ─────────────────────────────────────────────────────────────

_global_context_engine: ContextEngine | None = None


def get_context_engine() -> ContextEngine:
    """Get or create the global context engine."""
    global _global_context_engine
    if _global_context_engine is None:
        _global_context_engine = ContextEngine()
    return _global_context_engine




"""Audit logging — every tool action gets recorded for review."""

from __future__ import annotations

import json
import time
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any

from plutus.config import plutus_dir


@dataclass
class AuditEntry:
    timestamp: float
    tool_name: str
    operation: str | None
    params: dict[str, Any]
    decision: str  # "allowed", "denied", "pending_approval", "approved", "rejected"
    tier: str
    reason: str
    result_summary: str | None = None
    id: str = field(default_factory=lambda: f"{time.time_ns()}")

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)


class AuditLogger:
    """Append-only audit log stored as newline-delimited JSON."""

    def __init__(self, path: Path | None = None):
        self._path = path or (plutus_dir() / "audit.jsonl")
        self._path.parent.mkdir(parents=True, exist_ok=True)

    def log(self, entry: AuditEntry) -> None:
        with open(self._path, "a") as f:
            f.write(json.dumps(entry.to_dict()) + "\n")

    def recent(self, limit: int = 50, offset: int = 0) -> list[AuditEntry]:
        """Read the most recent audit entries."""
        if not self._path.exists():
            return []

        lines = self._path.read_text().strip().split("\n")
        lines = [l for l in lines if l]  # skip blanks
        lines.reverse()  # newest first

        entries = []
        for line in lines[offset : offset + limit]:
            data = json.loads(line)
            entries.append(AuditEntry(**data))
        return entries

    def count(self) -> int:
        if not self._path.exists():
            return 0
        return sum(1 for line in self._path.read_text().strip().split("\n") if line)

    def clear(self) -> None:
        """Wipe the audit log."""
        self._path.write_text("")

    def entries_for_tool(self, tool_name: str, limit: int = 20) -> list[AuditEntry]:
        entries = self.recent(limit=500)
        return [e for e in entries if e.tool_name == tool_name][:limit]

    def summary(self) -> dict[str, Any]:
        """Return a summary of audit activity."""
        entries = self.recent(limit=1000)
        by_decision: dict[str, int] = {}
        by_tool: dict[str, int] = {}
        for e in entries:
            by_decision[e.decision] = by_decision.get(e.decision, 0) + 1
            by_tool[e.tool_name] = by_tool.get(e.tool_name, 0) + 1

        return {
            "total_entries": len(entries),
            "by_decision": by_decision,
            "by_tool": by_tool,
            "latest": entries[0].to_dict() if entries else None,
        }



"""Filesystem tool — read, write, search, and manage files."""

from __future__ import annotations

import glob
import logging
import os
from pathlib import Path
from typing import Any

from plutus.tools.base import Tool

logger = logging.getLogger("plutus.tools.filesystem")

MAX_FILE_SIZE = 1_000_000  # 1MB read limit
MAX_READ_LINES = 2000


class FilesystemTool(Tool):
    @property
    def name(self) -> str:
        return "filesystem"

    @property
    def description(self) -> str:
        return (
            "Interact with the local filesystem. Read, write, search, list, and manage files "
            "and directories. Supports operations: read, write, append, list, search, "
            "mkdir, delete, move, copy, info."
        )

    @property
    def parameters(self) -> dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "operation": {
                    "type": "string",
                    "enum": [
                        "read", "write", "append", "list", "search",
                        "mkdir", "delete", "move", "copy", "info",
                    ],
                    "description": "The filesystem operation to perform",
                },
                "path": {
                    "type": "string",
                    "description": "Target file or directory path",
                },
                "content": {
                    "type": "string",
                    "description": "Content to write (for write/append operations)",
                },
                "destination": {
                    "type": "string",
                    "description": "Destination path (for move/copy operations)",
                },
                "pattern": {
                    "type": "string",
                    "description": "Glob or search pattern (for search/list operations)",
                },
                "recursive": {
                    "type": "boolean",
                    "description": "Recurse into subdirectories (default: false)",
                },
            },
            "required": ["operation", "path"],
        }

    async def execute(self, **kwargs: Any) -> str:
        operation: str = kwargs["operation"]
        path: str = kwargs["path"]

        # Expand ~ and resolve
        resolved = Path(path).expanduser().resolve()

        handlers = {
            "read": self._read,
            "write": self._write,
            "append": self._append,
            "list": self._list,
            "search": self._search,
            "mkdir": self._mkdir,
            "delete": self._delete,
            "move": self._move,
            "copy": self._copy,
            "info": self._info,
        }

        handler = handlers.get(operation)
        if not handler:
            return f"[ERROR] Unknown operation: {operation}"

        try:
            return await handler(resolved, kwargs)
        except PermissionError:
            return f"[ERROR] Permission denied: {resolved}"
        except FileNotFoundError:
            return f"[ERROR] File not found: {resolved}"
        except Exception as e:
            return f"[ERROR] {e}"

    async def _read(self, path: Path, kwargs: dict) -> str:
        if not path.exists():
            return f"[ERROR] File not found: {path}"
        if path.stat().st_size > MAX_FILE_SIZE:
            return f"[ERROR] File too large ({path.stat().st_size} bytes). Max: {MAX_FILE_SIZE}"

        content = path.read_text(errors="replace")
        lines = content.split("\n")
        if len(lines) > MAX_READ_LINES:
            content = "\n".join(lines[:MAX_READ_LINES])
            content += f"\n... [{len(lines) - MAX_READ_LINES} more lines]"
        return content

    async def _write(self, path: Path, kwargs: dict) -> str:
        content = kwargs.get("content")
        if content is None:
            # Check for common alternative key names the LLM might use
            content = kwargs.get("text") or kwargs.get("data") or kwargs.get("file_content")
        if content is None:
            logger.error(f"filesystem._write called with no content! kwargs keys: {list(kwargs.keys())}")
            return (
                "[ERROR] No 'content' parameter provided. "
                "You MUST pass content='your file content here' to write a file. "
                f"Received parameters: {list(kwargs.keys())}"
            )
        if not content:
            logger.warning(f"filesystem._write called with empty content string for {path}")
            return (
                "[WARNING] Content is an empty string — this will create an empty file. "
                "If this is unintentional, call filesystem again with the actual content."
            )
        logger.info(f"filesystem._write: writing {len(content)} chars to {path}")
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(content, encoding="utf-8")
        return f"Successfully written {len(content)} characters ({len(content.encode('utf-8'))} bytes) to {path}"

    async def _append(self, path: Path, kwargs: dict) -> str:
        content = kwargs.get("content")
        if content is None:
            content = kwargs.get("text") or kwargs.get("data") or kwargs.get("file_content")
        if content is None:
            logger.error(f"filesystem._append called with no content! kwargs keys: {list(kwargs.keys())}")
            return (
                "[ERROR] No 'content' parameter provided. "
                "You MUST pass content='text to append' to append to a file. "
                f"Received parameters: {list(kwargs.keys())}"
            )
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "a", encoding="utf-8") as f:
            f.write(content)
        return f"Appended {len(content)} characters to {path}"

    async def _list(self, path: Path, kwargs: dict) -> str:
        if not path.is_dir():
            return f"[ERROR] Not a directory: {path}"

        pattern = kwargs.get("pattern", "*")
        recursive = kwargs.get("recursive", False)

        if recursive:
            entries = sorted(path.rglob(pattern))
        else:
            entries = sorted(path.glob(pattern))

        if not entries:
            return f"No entries matching '{pattern}' in {path}"

        lines = []
        for entry in entries[:200]:  # limit output
            prefix = "d" if entry.is_dir() else "f"
            size = entry.stat().st_size if entry.is_file() else 0
            lines.append(f"[{prefix}] {entry.relative_to(path)}  ({size} bytes)")

        result = "\n".join(lines)
        if len(entries) > 200:
            result += f"\n... [{len(entries) - 200} more entries]"
        return result

    async def _search(self, path: Path, kwargs: dict) -> str:
        pattern = kwargs.get("pattern", "")
        if not pattern:
            return "[ERROR] Search requires a 'pattern' parameter"

        results = []
        search_path = path if path.is_dir() else path.parent

        for file_path in search_path.rglob("*"):
            if not file_path.is_file():
                continue
            if file_path.stat().st_size > MAX_FILE_SIZE:
                continue
            try:
                content = file_path.read_text(errors="replace")
                for i, line in enumerate(content.split("\n"), 1):
                    if pattern.lower() in line.lower():
                        results.append(f"{file_path}:{i}: {line.strip()}")
                        if len(results) >= 50:
                            return "\n".join(results) + "\n... [more results truncated]"
            except (PermissionError, UnicodeDecodeError):
                continue

        return "\n".join(results) if results else f"No matches for '{pattern}'"

    async def _mkdir(self, path: Path, kwargs: dict) -> str:
        path.mkdir(parents=True, exist_ok=True)
        return f"Created directory: {path}"

    async def _delete(self, path: Path, kwargs: dict) -> str:
        if path.is_file():
            path.unlink()
            return f"Deleted file: {path}"
        elif path.is_dir():
            import shutil
            shutil.rmtree(path)
            return f"Deleted directory: {path}"
        return f"[ERROR] Path does not exist: {path}"

    async def _move(self, path: Path, kwargs: dict) -> str:
        dest = kwargs.get("destination")
        if not dest:
            return "[ERROR] Move requires a 'destination' parameter"
        dest_path = Path(dest).expanduser().resolve()
        path.rename(dest_path)
        return f"Moved {path} → {dest_path}"

    async def _copy(self, path: Path, kwargs: dict) -> str:
        import shutil
        dest = kwargs.get("destination")
        if not dest:
            return "[ERROR] Copy requires a 'destination' parameter"
        dest_path = Path(dest).expanduser().resolve()
        if path.is_file():
            shutil.copy2(path, dest_path)
        else:
            shutil.copytree(path, dest_path)
        return f"Copied {path} → {dest_path}"

    async def _info(self, path: Path, kwargs: dict) -> str:
        if not path.exists():
            return f"[ERROR] Path does not exist: {path}"
        stat = path.stat()
        return (
            f"path: {path}\n"
            f"type: {'directory' if path.is_dir() else 'file'}\n"
            f"size: {stat.st_size} bytes\n"
            f"permissions: {oct(stat.st_mode)}\n"
            f"modified: {stat.st_mtime}\n"
        )



"""Policy evaluation — decides whether a tool action is allowed, denied, or needs approval."""

from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any

from plutus.guardrails.tiers import (
    TIER_DEFAULTS,
    Tier,
    ToolPermission,
    ToolPolicy,
    default_permission_for_tier,
)


@dataclass
class PolicyDecision:
    """Result of evaluating a tool action against the current policy."""

    allowed: bool
    requires_approval: bool
    reason: str
    tool_name: str
    operation: str | None = None

    @property
    def status(self) -> str:
        if self.allowed and not self.requires_approval:
            return "allowed"
        if self.requires_approval:
            return "pending_approval"
        return "denied"


class PolicyEvaluator:
    """Evaluates tool actions against the current tier and any user overrides."""

    def __init__(self, tier: Tier, overrides: dict[str, dict[str, Any]] | None = None):
        self.tier = tier
        self._overrides = overrides or {}

    def evaluate(
        self,
        tool_name: str,
        operation: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> PolicyDecision:
        """Check whether a tool action is allowed under the current policy.

        Args:
            tool_name: The tool being invoked (e.g. "shell", "filesystem").
            operation: The specific operation (e.g. "read", "write", "execute").
            params: Action parameters for pattern matching (e.g. {"command": "ls -la"}).
        """
        # Check user-level overrides first
        if tool_name in self._overrides:
            override = self._overrides[tool_name]
            if not override.get("enabled", True):
                return PolicyDecision(
                    allowed=False,
                    requires_approval=False,
                    reason=f"Tool '{tool_name}' is disabled by user configuration",
                    tool_name=tool_name,
                    operation=operation,
                )
            if override.get("require_approval"):
                return PolicyDecision(
                    allowed=True,
                    requires_approval=True,
                    reason=f"Tool '{tool_name}' requires approval per user override",
                    tool_name=tool_name,
                    operation=operation,
                )

        # Look up tier defaults
        tier_policies = TIER_DEFAULTS.get(self.tier, {})
        policy = tier_policies.get(tool_name)

        if policy is None:
            # Tool not explicitly listed — use tier default
            fallback = default_permission_for_tier(self.tier)
            return self._decision_from_permission(fallback, tool_name, operation, "tier default")

        # Check denied patterns (blocklist)
        if policy.denied_patterns and params:
            for pattern in policy.denied_patterns:
                for val in params.values():
                    if isinstance(val, str) and re.search(re.escape(pattern), val):
                        return PolicyDecision(
                            allowed=False,
                            requires_approval=False,
                            reason=f"Action matches denied pattern: '{pattern}'",
                            tool_name=tool_name,
                            operation=operation,
                        )

        # Check operation allowlist
        if policy.allowed_operations and operation:
            if operation not in policy.allowed_operations:
                return PolicyDecision(
                    allowed=False,
                    requires_approval=False,
                    reason=(
                        f"Operation '{operation}' not in allowed list: "
                        f"{policy.allowed_operations}"
                    ),
                    tool_name=tool_name,
                    operation=operation,
                )

        return self._decision_from_permission(policy.permission, tool_name, operation, "tier policy")

    def _decision_from_permission(
        self,
        permission: ToolPermission,
        tool_name: str,
        operation: str | None,
        source: str,
    ) -> PolicyDecision:
        if permission == ToolPermission.ALLOWED:
            return PolicyDecision(
                allowed=True,
                requires_approval=False,
                reason=f"Allowed by {source}",
                tool_name=tool_name,
                operation=operation,
            )
        if permission == ToolPermission.REQUIRES_APPROVAL:
            return PolicyDecision(
                allowed=True,
                requires_approval=True,
                reason=f"Requires approval per {source}",
                tool_name=tool_name,
                operation=operation,
            )
        return PolicyDecision(
            allowed=False,
            requires_approval=False,
            reason=f"Denied by {source}",
            tool_name=tool_name,
            operation=operation,
        )



"""WSL tool — execute commands inside Windows Subsystem for Linux.

Gives Plutus fluent access to a full Linux environment from Windows, including
package management, shell scripting, Linux-native CLI tools, and cross-OS
file operations.  On non-Windows hosts the tool falls back to the native shell
so prompts that target "the Linux side" still work everywhere.
"""

from __future__ import annotations

import asyncio
import os
import platform
import shutil
from typing import Any

from plutus.tools.base import Tool

# Hard-blocked regardless of guardrail tier
BLOCKED_PATTERNS = [
    "rm -rf /",
    "mkfs.",
    "dd if=/dev/zero",
    ":(){ :|:& };:",
    "> /dev/sda",
    "chmod -R 777 /",
]

MAX_OUTPUT_LENGTH = 50_000


def _wsl_available() -> bool:
    """Return True when a usable WSL installation is detected."""
    if platform.system() != "Windows":
        return False
    return shutil.which("wsl") is not None or shutil.which("wsl.exe") is not None


class WSLTool(Tool):
    """Execute commands inside WSL or the native Linux/macOS shell.

    On **Windows** the tool routes commands through ``wsl.exe`` so the agent
    can leverage the full Linux toolchain (apt, grep, sed, awk, ssh, docker,
    …).  On **Linux / macOS** it runs commands natively, so prompts that ask
    for "Linux" behaviour work on every platform.
    """

    # ── Tool interface ────────────────────────────────────────

    @property
    def name(self) -> str:
        return "wsl"

    @property
    def description(self) -> str:
        host = platform.system()
        if host == "Windows":
            return (
                "Execute commands inside the Windows Subsystem for Linux (WSL). "
                "Use this when you need Linux tools, package managers (apt, pip), "
                "shell scripting, SSH, Docker, compilers, or any Linux-native CLI "
                "utility. Supports choosing a specific WSL distro, setting the "
                "working directory, managing distros, and translating paths between "
                "Windows and Linux. "
                "The command string is passed directly to bash — write normal Linux "
                "commands with quotes, pipes, and redirects as-is (no Windows escaping needed)."
            )
        return (
            "Execute Linux / Unix shell commands. "
            "Use this for package management (apt, brew, pip), shell scripting, "
            "SSH, Docker, compilers, and any Unix-native CLI utility."
        )

    @property
    def parameters(self) -> dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "operation": {
                    "type": "string",
                    "enum": [
                        "run",
                        "list_distros",
                        "set_default",
                        "path_to_linux",
                        "path_to_windows",
                        "info",
                    ],
                    "description": (
                        "Operation to perform. "
                        "'run' — execute a command; "
                        "'list_distros' — list installed WSL distributions; "
                        "'set_default' — change the default distro; "
                        "'path_to_linux' — convert a Windows path to its /mnt/… equivalent; "
                        "'path_to_windows' — convert a Linux /mnt/… path to Windows; "
                        "'info' — show WSL version, default distro, and status."
                    ),
                },
                "command": {
                    "type": "string",
                    "description": "Shell command to execute (required for 'run').",
                },
                "distro": {
                    "type": "string",
                    "description": (
                        "Target WSL distribution name (e.g. 'Ubuntu', 'Debian'). "
                        "If omitted the default distro is used."
                    ),
                },
                "working_directory": {
                    "type": "string",
                    "description": (
                        "Working directory *inside* WSL / the Linux shell. "
                        "Defaults to the user's home directory."
                    ),
                },
                "path": {
                    "type": "string",
                    "description": "Path to translate (for path_to_linux / path_to_windows).",
                },
                "timeout": {
                    "type": "integer",
                    "description": "Timeout in seconds (default: 120).",
                },
                "user": {
                    "type": "string",
                    "description": "Run the command as this Linux user (WSL -u flag).",
                },
            },
            "required": ["operation"],
        }

    # ── Execution dispatch ────────────────────────────────────

    async def execute(self, **kwargs: Any) -> str:
        operation: str = kwargs.get("operation", "run")

        handlers = {
            "run": self._run,
            "list_distros": self._list_distros,
            "set_default": self._set_default,
            "path_to_linux": self._path_to_linux,
            "path_to_windows": self._path_to_windows,
            "info": self._info,
        }

        handler = handlers.get(operation)
        if handler is None:
            return f"[ERROR] Unknown operation: {operation}"

        try:
            return await handler(kwargs)
        except Exception as e:
            return f"[ERROR] {e}"

    # ── Operations ────────────────────────────────────────────

    async def _run(self, kwargs: dict[str, Any]) -> str:
        command: str | None = kwargs.get("command")
        if not command:
            return "[ERROR] 'command' is required for the 'run' operation."

        # Safety check
        cmd_lower = command.lower().strip()
        for blocked in BLOCKED_PATTERNS:
            if blocked in cmd_lower:
                return f"[BLOCKED] Command contains blocked pattern: '{blocked}'"

        distro: str | None = kwargs.get("distro")
        user: str | None = kwargs.get("user")
        working_dir: str | None = kwargs.get("working_directory")
        timeout: int = kwargs.get("timeout", 120)

        if platform.system() == "Windows" and _wsl_available():
            # Use argv list so CMD.exe never interprets the command string.
            argv = self._build_wsl_argv(
                command, distro=distro, user=user, cwd=working_dir,
            )
            return await self._exec(argv, timeout=timeout)

        # Native Linux / macOS — run through the shell directly.
        shell_cmd = self._build_native_shell_cmd(command, cwd=working_dir)
        cwd = working_dir or os.path.expanduser("~")
        return await self._exec(shell_cmd, cwd=cwd, timeout=timeout)

    async def _list_distros(self, _kwargs: dict[str, Any]) -> str:
        if platform.system() != "Windows":
            return (
                "platform: Linux / macOS (native)\n"
                "WSL is not applicable — commands already run in a native Unix shell."
            )

        if not _wsl_available():
            return "[ERROR] WSL is not installed or not found on PATH."

        # wsl --list --verbose gives name, state, and version
        return await self._exec(["wsl.exe", "--list", "--verbose"], timeout=15)

    async def _set_default(self, kwargs: dict[str, Any]) -> str:
        distro: str | None = kwargs.get("distro")
        if not distro:
            return "[ERROR] 'distro' is required for set_default."

        if platform.system() != "Windows":
            return "[ERROR] set_default is only available on Windows with WSL."

        if not _wsl_available():
            return "[ERROR] WSL is not installed or not found on PATH."

        return await self._exec(["wsl.exe", "--set-default", distro], timeout=15)

    async def _path_to_linux(self, kwargs: dict[str, Any]) -> str:
        path: str | None = kwargs.get("path")
        if not path:
            return "[ERROR] 'path' is required."

        if platform.system() != "Windows":
            return f"linux_path: {path}  (already on a Unix system)"

        if not _wsl_available():
            return "[ERROR] WSL is not installed or not found on PATH."

        result = await self._exec(["wsl.exe", "wslpath", "-u", path], timeout=10)
        return f"linux_path: {result.strip()}" if "[" not in result else result

    async def _path_to_windows(self, kwargs: dict[str, Any]) -> str:
        path: str | None = kwargs.get("path")
        if not path:
            return "[ERROR] 'path' is required."

        if platform.system() != "Windows":
            return f"windows_path: (not applicable — running on {platform.system()})"

        if not _wsl_available():
            return "[ERROR] WSL is not installed or not found on PATH."

        result = await self._exec(["wsl.exe", "wslpath", "-w", path], timeout=10)
        return f"windows_path: {result.strip()}" if "[" not in result else result

    async def _info(self, _kwargs: dict[str, Any]) -> str:
        host = platform.system()
        parts: list[str] = [f"host_os: {host}"]

        if host == "Windows":
            if not _wsl_available():
                parts.append("wsl_status: not installed")
                return "\n".join(parts)

            # WSL version
            ver = await self._exec(["wsl.exe", "--version"], timeout=10)
            parts.append(f"wsl_version:\n{ver.strip()}")

            # Default distro
            distros = await self._exec(["wsl.exe", "--list", "--verbose"], timeout=10)
            parts.append(f"distros:\n{distros.strip()}")
        else:
            import shutil as _shutil

            shell = os.environ.get("SHELL", "unknown")
            parts.append(f"shell: {shell}")
            parts.append(f"wsl_status: not applicable (native {host})")
            # Useful info for the agent to know what's available
            for cmd in ("apt", "brew", "dnf", "pacman", "docker", "ssh", "python3", "node", "gcc"):
                loc = _shutil.which(cmd)
                if loc:
                    parts.append(f"  {cmd}: {loc}")

        return "\n".join(parts)

    # ── Helpers ───────────────────────────────────────────────

    @staticmethod
    def _build_wsl_argv(
        command: str,
        *,
        distro: str | None = None,
        user: str | None = None,
        cwd: str | None = None,
    ) -> list[str]:
        """Build an argument list for ``wsl.exe`` that bypasses CMD parsing.

        Returns a list of strings suitable for ``create_subprocess_exec`` so
        that Windows CMD.exe never interprets quotes, pipes, redirects, or
        heredocs in *command*.
        """
        argv = ["wsl.exe"]
        if distro:
            argv += ["-d", distro]
        if user:
            argv += ["-u", user]
        if cwd:
            argv += ["--cd", cwd]
        # Pass the command to bash -ic as a single token.
        # Because we use create_subprocess_exec (no shell), the command
        # string is handed to wsl.exe → bash verbatim — no CMD mangling.
        argv += ["--", "bash", "-ic", command]
        return argv

    @staticmethod
    def _build_native_shell_cmd(
        command: str,
        *,
        cwd: str | None = None,
    ) -> str:
        """Build a shell command string for native Linux/macOS execution."""
        if cwd:
            return f"cd {_shell_quote(cwd)} && {command}"
        return command

    @staticmethod
    async def _exec(
        cmd: str | list[str], *, cwd: str | None = None, timeout: int = 120
    ) -> str:
        """Run *cmd* asynchronously and return formatted output.

        *cmd* can be a string (passed to the shell) or a list of arguments
        (executed directly, bypassing the shell).  On Windows the list form
        **must** be used for WSL commands so that CMD.exe does not mangle
        quotes, pipes, or special characters.
        """
        try:
            if isinstance(cmd, list):
                process = await asyncio.create_subprocess_exec(
                    *cmd,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                    cwd=cwd,
                )
            else:
                process = await asyncio.create_subprocess_shell(
                    cmd,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                    cwd=cwd,
                )

            stdout, stderr = await asyncio.wait_for(
                process.communicate(), timeout=timeout
            )

            stdout_text = stdout.decode("utf-8", errors="replace")
            stderr_text = stderr.decode("utf-8", errors="replace")

            if len(stdout_text) > MAX_OUTPUT_LENGTH:
                stdout_text = stdout_text[:MAX_OUTPUT_LENGTH] + "\n... [truncated]"
            if len(stderr_text) > MAX_OUTPUT_LENGTH:
                stderr_text = stderr_text[:MAX_OUTPUT_LENGTH] + "\n... [truncated]"

            result_parts: list[str] = []
            if stdout_text.strip():
                result_parts.append(f"stdout:\n{stdout_text.strip()}")
            if stderr_text.strip():
                result_parts.append(f"stderr:\n{stderr_text.strip()}")
            result_parts.append(f"exit_code: {process.returncode}")

            return "\n".join(result_parts)

        except TimeoutError:
            return f"[TIMEOUT] Command timed out after {timeout} seconds"
        except Exception as e:
            return f"[ERROR] {e}"


def _shell_quote(s: str) -> str:
    """Simple POSIX-style single-quote wrapper."""
    return "'" + s.replace("'", "'\\''") + "'"



"""Scheduler Tool — allows the agent to create and manage scheduled tasks.

Operations:
  - create:      Create a new scheduled job (cron, interval, or one-shot)
  - list:        List all scheduled jobs
  - get:         Get details of a specific job
  - update:      Update a job's settings
  - delete:      Delete a scheduled job
  - pause:       Pause a job
  - resume:      Resume a paused job
  - history:     View recent job execution history
  - stats:       Get scheduler statistics
"""

from __future__ import annotations

import json
import time
from typing import Any

from plutus.core.scheduler import (
    Scheduler,
    ScheduledJob,
    JobType,
    JobState,
)
from plutus.tools.base import Tool


class SchedulerTool(Tool):
    """Create and manage scheduled tasks (cron jobs, intervals, one-shots)."""

    def __init__(self, scheduler: Scheduler):
        self._scheduler = scheduler

    @property
    def name(self) -> str:
        return "scheduler"

    @property
    def description(self) -> str:
        return (
            "Create and manage scheduled tasks. Supports cron expressions, intervals, and one-shot timers.\n\n"
            "Operations:\n"
            "- create: Create a new scheduled job\n"
            "- list: List all scheduled jobs\n"
            "- get: Get details of a specific job\n"
            "- update: Update a job's settings\n"
            "- delete: Delete a scheduled job\n"
            "- pause/resume: Pause or resume a job\n"
            "- history: View recent execution history\n"
            "- stats: Get scheduler statistics\n\n"
            "Schedule types:\n"
            "- cron: Standard 5-field cron expression (min hour dom month dow)\n"
            "  Examples: '0 6 * * *' (daily 6AM), '*/5 * * * *' (every 5 min), '0 9 * * 1-5' (weekdays 9AM)\n"
            "- interval: Run every N seconds (e.g., 300 for every 5 minutes)\n"
            "- once: Run once at a specific time\n\n"
            "Each job can spawn a worker (default) or run on the main agent. "
            "Set model_key to choose which model handles the job."
        )

    @property
    def parameters(self) -> dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "operation": {
                    "type": "string",
                    "enum": ["create", "list", "get", "update", "delete", "pause", "resume", "history", "stats"],
                    "description": "The scheduler operation to perform.",
                },
                "name": {
                    "type": "string",
                    "description": "Human-readable name for the job (for 'create').",
                },
                "description": {
                    "type": "string",
                    "description": "Description of what the job does (for 'create').",
                },
                "job_type": {
                    "type": "string",
                    "enum": ["cron", "interval", "once"],
                    "description": "Type of schedule (for 'create'). Default: 'cron'.",
                },
                "schedule": {
                    "type": "string",
                    "description": "Cron expression for 'cron' type (e.g., '0 6 * * *' for daily at 6AM).",
                },
                "interval_seconds": {
                    "type": "integer",
                    "description": "Interval in seconds for 'interval' type (e.g., 300 for every 5 minutes).",
                },
                "prompt": {
                    "type": "string",
                    "description": "The instruction to execute when the job fires.",
                },
                "model_key": {
                    "type": "string",
                    "enum": ["claude-haiku", "claude-sonnet", "claude-opus", "gpt-5.2"],
                    "description": "Which model to use for this job. Omit for auto-selection.",
                },
                "spawn_worker": {
                    "type": "boolean",
                    "description": "Whether to spawn a worker (true, default) or run on main agent (false).",
                },
                "max_runs": {
                    "type": "integer",
                    "description": "Maximum number of times to run (0 = unlimited). Default: 0.",
                },
                "job_id": {
                    "type": "string",
                    "description": "Job ID (for 'get', 'update', 'delete', 'pause', 'resume', 'history').",
                },
                "updates": {
                    "type": "object",
                    "description": "Fields to update (for 'update' operation).",
                },
                "limit": {
                    "type": "integer",
                    "description": "Max results to return (for 'history'). Default: 20.",
                },
            },
            "required": ["operation"],
        }

    async def execute(self, **kwargs: Any) -> Any:
        operation = kwargs.get("operation", "list")

        try:
            if operation == "create":
                return self._create(kwargs)
            elif operation == "list":
                return self._list()
            elif operation == "get":
                return self._get(kwargs)
            elif operation == "update":
                return self._update(kwargs)
            elif operation == "delete":
                return self._delete(kwargs)
            elif operation == "pause":
                return self._pause(kwargs)
            elif operation == "resume":
                return self._resume(kwargs)
            elif operation == "history":
                return self._history(kwargs)
            elif operation == "stats":
                return self._stats()
            else:
                return f"[ERROR] Unknown operation: {operation}"
        except Exception as e:
            return f"[ERROR] Scheduler operation failed: {e}"

    def _create(self, kwargs: dict) -> str:
        prompt = kwargs.get("prompt", "")
        if not prompt:
            return "[ERROR] 'prompt' is required — what should the job do?"

        job_type_str = kwargs.get("job_type", "cron")
        try:
            job_type = JobType(job_type_str)
        except ValueError:
            return f"[ERROR] Invalid job_type: {job_type_str}. Use 'cron', 'interval', or 'once'."

        schedule = kwargs.get("schedule", "")
        interval_seconds = kwargs.get("interval_seconds", 0)

        if job_type == JobType.CRON and not schedule:
            return "[ERROR] 'schedule' (cron expression) is required for cron jobs."
        if job_type == JobType.INTERVAL and not interval_seconds:
            return "[ERROR] 'interval_seconds' is required for interval jobs."

        job = ScheduledJob(
            name=kwargs.get("name", "Unnamed Job"),
            description=kwargs.get("description", ""),
            job_type=job_type,
            schedule=schedule,
            interval_seconds=interval_seconds,
            prompt=prompt,
            model_key=kwargs.get("model_key"),
            spawn_worker=kwargs.get("spawn_worker", True),
            max_runs=kwargs.get("max_runs", 0),
        )

        created = self._scheduler.add_job(job)
        return json.dumps({
            "success": True,
            "message": f"Job '{created.name}' created successfully.",
            "job": created.to_dict(),
        }, indent=2)

    def _list(self) -> str:
        jobs = self._scheduler.list_jobs()
        if not jobs:
            return json.dumps({"jobs": [], "message": "No scheduled jobs."})
        return json.dumps({"jobs": jobs, "total": len(jobs)}, indent=2)

    def _get(self, kwargs: dict) -> str:
        job_id = kwargs.get("job_id", "")
        if not job_id:
            return "[ERROR] 'job_id' is required."
        job = self._scheduler.get_job(job_id)
        if not job:
            return f"[ERROR] Job '{job_id}' not found."
        return json.dumps(job.to_dict(), indent=2)

    def _update(self, kwargs: dict) -> str:
        job_id = kwargs.get("job_id", "")
        updates = kwargs.get("updates", {})
        if not job_id:
            return "[ERROR] 'job_id' is required."
        if not updates:
            return "[ERROR] 'updates' dict is required."
        job = self._scheduler.update_job(job_id, updates)
        if not job:
            return f"[ERROR] Job '{job_id}' not found."
        return json.dumps({
            "success": True,
            "message": f"Job '{job.name}' updated.",
            "job": job.to_dict(),
        }, indent=2)

    def _delete(self, kwargs: dict) -> str:
        job_id = kwargs.get("job_id", "")
        if not job_id:
            return "[ERROR] 'job_id' is required."
        success = self._scheduler.remove_job(job_id)
        if success:
            return f"Job '{job_id}' deleted."
        return f"[ERROR] Job '{job_id}' not found."

    def _pause(self, kwargs: dict) -> str:
        job_id = kwargs.get("job_id", "")
        if not job_id:
            return "[ERROR] 'job_id' is required."
        success = self._scheduler.pause_job(job_id)
        if success:
            return f"Job '{job_id}' paused."
        return f"[ERROR] Job '{job_id}' not found."

    def _resume(self, kwargs: dict) -> str:
        job_id = kwargs.get("job_id", "")
        if not job_id:
            return "[ERROR] 'job_id' is required."
        success = self._scheduler.resume_job(job_id)
        if success:
            return f"Job '{job_id}' resumed."
        return f"[ERROR] Job '{job_id}' not found."

    def _history(self, kwargs: dict) -> str:
        job_id = kwargs.get("job_id")
        limit = kwargs.get("limit", 20)
        execs = self._scheduler.list_executions(limit=limit, job_id=job_id)
        return json.dumps({"executions": execs, "total": len(execs)}, indent=2)

    def _stats(self) -> str:
        return json.dumps(self._scheduler.stats(), indent=2)



"""Clipboard tool — read and write system clipboard."""

from __future__ import annotations

import asyncio
import platform
from typing import Any

from plutus.tools.base import Tool


class ClipboardTool(Tool):
    @property
    def name(self) -> str:
        return "clipboard"

    @property
    def description(self) -> str:
        return "Read from or write to the system clipboard. Operations: read, write."

    @property
    def parameters(self) -> dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "operation": {
                    "type": "string",
                    "enum": ["read", "write"],
                    "description": "Read from or write to the clipboard",
                },
                "content": {
                    "type": "string",
                    "description": "Content to write to clipboard (for write operation)",
                },
            },
            "required": ["operation"],
        }

    async def execute(self, **kwargs: Any) -> str:
        operation: str = kwargs["operation"]

        if operation == "read":
            return await self._read()
        elif operation == "write":
            content = kwargs.get("content", "")
            return await self._write(content)
        return f"[ERROR] Unknown operation: {operation}"

    async def _read(self) -> str:
        system = platform.system()
        try:
            if system == "Darwin":
                cmd = "pbpaste"
            elif system == "Linux":
                cmd = "xclip -selection clipboard -o"
            elif system == "Windows":
                cmd = "powershell Get-Clipboard"
            else:
                return "[ERROR] Unsupported platform for clipboard"

            proc = await asyncio.create_subprocess_shell(
                cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
            )
            stdout, stderr = await proc.communicate()
            if proc.returncode != 0:
                return f"[ERROR] Clipboard read failed: {stderr.decode()}"
            return stdout.decode("utf-8", errors="replace")
        except Exception as e:
            return f"[ERROR] {e}"

    async def _write(self, content: str) -> str:
        system = platform.system()
        try:
            if system == "Darwin":
                cmd = "pbcopy"
            elif system == "Linux":
                cmd = "xclip -selection clipboard"
            elif system == "Windows":
                cmd = "powershell Set-Clipboard"
            else:
                return "[ERROR] Unsupported platform for clipboard"

            proc = await asyncio.create_subprocess_shell(
                cmd,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            _, stderr = await proc.communicate(content.encode())
            if proc.returncode != 0:
                return f"[ERROR] Clipboard write failed: {stderr.decode()}"
            return f"Written {len(content)} characters to clipboard"
        except Exception as e:
            return f"[ERROR] {e}"



"""Window manager — deep cross-platform window orchestration.

Provides intelligent window management: find windows by title/app,
snap to screen edges, tile multiple windows, switch between them,
and manage virtual desktops. Works on Windows, macOS, and Linux.
"""

from __future__ import annotations

import asyncio
import platform
import re
import shutil
from dataclasses import dataclass
from typing import Any

SYSTEM = platform.system()


@dataclass
class WindowInfo:
    """Information about a desktop window."""
    id: str = ""
    title: str = ""
    app: str = ""
    x: int = 0
    y: int = 0
    width: int = 0
    height: int = 0
    is_active: bool = False
    is_minimized: bool = False
    is_maximized: bool = False
    pid: int = 0

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "title": self.title,
            "app": self.app,
            "x": self.x,
            "y": self.y,
            "width": self.width,
            "height": self.height,
            "is_active": self.is_active,
            "is_minimized": self.is_minimized,
            "is_maximized": self.is_maximized,
            "pid": self.pid,
        }


class WindowManager:
    """Cross-platform window manager with intelligent operations.

    Usage:
        wm = WindowManager()
        windows = await wm.list_windows()
        await wm.focus("Chrome")
        await wm.snap_left("Chrome")
        await wm.snap_right("Code")
        await wm.tile_windows(["Chrome", "Code", "Terminal"])
        await wm.close("Notepad")
    """

    async def _run(self, cmd: str, timeout: float = 10) -> tuple[str, str, int]:
        """Run a shell command and return (stdout, stderr, returncode)."""
        proc = await asyncio.create_subprocess_shell(
            cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        try:
            stdout, stderr = await asyncio.wait_for(proc.communicate(), timeout=timeout)
            return (
                stdout.decode("utf-8", errors="replace").strip(),
                stderr.decode("utf-8", errors="replace").strip(),
                proc.returncode or 0,
            )
        except asyncio.TimeoutError:
            proc.kill()
            return ("", "Command timed out", 1)

    async def _get_screen_size(self) -> tuple[int, int]:
        """Get screen dimensions."""
        if SYSTEM == "Windows":
            out, _, _ = await self._run(
                'powershell -Command "[System.Windows.Forms.Screen]::PrimaryScreen.Bounds | '
                'Select-Object Width,Height | ConvertTo-Json"'
            )
            try:
                import json
                data = json.loads(out)
                return (data["Width"], data["Height"])
            except Exception:
                return (1920, 1080)
        elif SYSTEM == "Darwin":
            out, _, _ = await self._run(
                "osascript -e 'tell application \"Finder\" to get bounds of window of desktop'"
            )
            try:
                parts = [int(x.strip()) for x in out.split(",")]
                return (parts[2], parts[3])
            except Exception:
                return (1920, 1080)
        else:
            # Use platform_utils for robust Linux detection (X11 + Wayland)
            try:
                from plutus.pc.platform_utils import get_screen_size_linux
                dims = get_screen_size_linux()
                if dims:
                    return dims
            except Exception:
                pass
            # Manual fallback
            out, _, _ = await self._run("xdpyinfo 2>/dev/null | grep dimensions")
            match = re.search(r'(\d+)x(\d+)', out)
            if match:
                return (int(match.group(1)), int(match.group(2)))
            out, _, _ = await self._run("xrandr 2>/dev/null | grep '*'")
            match = re.search(r'(\d+)x(\d+)', out)
            if match:
                return (int(match.group(1)), int(match.group(2)))
            return (1920, 1080)

    # ─── Window Listing ───

    async def list_windows(self) -> list[dict[str, Any]]:
        """List all visible windows with details."""
        if SYSTEM == "Windows":
            return await self._list_windows_windows()
        elif SYSTEM == "Darwin":
            return await self._list_windows_macos()
        else:
            return await self._list_windows_linux()

    async def _list_windows_windows(self) -> list[dict[str, Any]]:
        cmd = (
            'powershell -Command "'
            "Get-Process | Where-Object {$_.MainWindowTitle -ne ''} | "
            "Select-Object Id, ProcessName, MainWindowTitle | ConvertTo-Json"
            '"'
        )
        out, _, _ = await self._run(cmd)
        windows = []
        try:
            import json
            data = json.loads(out)
            if isinstance(data, dict):
                data = [data]
            for proc in data:
                windows.append(WindowInfo(
                    id=str(proc.get("Id", "")),
                    title=proc.get("MainWindowTitle", ""),
                    app=proc.get("ProcessName", ""),
                    pid=proc.get("Id", 0),
                ).to_dict())
        except Exception:
            pass
        return windows

    async def _list_windows_macos(self) -> list[dict[str, Any]]:
        script = '''
        osascript -e '
        set output to ""
        tell application "System Events"
            set procs to every process whose visible is true
            repeat with p in procs
                set pName to name of p
                try
                    set wins to windows of p
                    repeat with w in wins
                        set wTitle to name of w
                        set output to output & pName & "|" & wTitle & "\\n"
                    end repeat
                end try
            end repeat
        end tell
        return output'
        '''
        out, _, _ = await self._run(script)
        windows = []
        for line in out.strip().split("\n"):
            if "|" in line:
                parts = line.split("|", 1)
                windows.append(WindowInfo(
                    app=parts[0].strip(),
                    title=parts[1].strip() if len(parts) > 1 else "",
                ).to_dict())
        return windows

    async def _list_windows_linux(self) -> list[dict[str, Any]]:
        windows = []

        # Try swaymsg first (Sway/Wayland compositor)
        if shutil.which("swaymsg"):
            try:
                out, _, rc = await self._run("swaymsg -t get_tree --raw")
                if rc == 0:
                    import json as _json
                    tree = _json.loads(out)
                    self._walk_sway_tree(tree, windows)
                    if windows:
                        return windows
            except Exception:
                pass

        if shutil.which("wmctrl"):
            out, _, _ = await self._run("wmctrl -l -p -G")
            for line in out.strip().split("\n"):
                if not line.strip():
                    continue
                parts = line.split(None, 8)
                if len(parts) >= 9:
                    windows.append(WindowInfo(
                        id=parts[0],
                        x=int(parts[2]) if parts[2].isdigit() else 0,
                        y=int(parts[3]) if parts[3].isdigit() else 0,
                        width=int(parts[4]) if parts[4].isdigit() else 0,
                        height=int(parts[5]) if parts[5].isdigit() else 0,
                        pid=int(parts[1]) if parts[1].isdigit() else 0,
                        title=parts[8] if len(parts) > 8 else "",
                    ).to_dict())
        elif shutil.which("xdotool"):
            out, _, _ = await self._run(
                "xdotool search --onlyvisible --name '' 2>/dev/null"
            )
            for wid in out.strip().split("\n"):
                wid = wid.strip()
                if not wid:
                    continue
                name_out, _, _ = await self._run(f"xdotool getwindowname {wid} 2>/dev/null")
                pid_out, _, _ = await self._run(f"xdotool getwindowpid {wid} 2>/dev/null")
                windows.append(WindowInfo(
                    id=wid,
                    title=name_out.strip(),
                    pid=int(pid_out.strip()) if pid_out.strip().isdigit() else 0,
                ).to_dict())
        return windows

    def _find_sway_focused(self, node: dict) -> dict | None:
        """Find the focused window in the Sway IPC tree."""
        if node.get("focused") and node.get("type") == "con" and node.get("name"):
            return node
        for child in node.get("nodes", []) + node.get("floating_nodes", []):
            result = self._find_sway_focused(child)
            if result:
                return result
        return None

    def _walk_sway_tree(self, node: dict, windows: list) -> None:
        """Recursively walk the Sway IPC tree to find windows."""
        if node.get("type") == "con" and node.get("name") and node.get("pid"):
            rect = node.get("rect", {})
            windows.append(WindowInfo(
                id=str(node.get("id", "")),
                title=node.get("name", ""),
                app=node.get("app_id", "") or node.get("window_properties", {}).get("class", ""),
                x=rect.get("x", 0),
                y=rect.get("y", 0),
                width=rect.get("width", 0),
                height=rect.get("height", 0),
                is_active=node.get("focused", False),
                pid=node.get("pid", 0),
            ).to_dict())
        for child in node.get("nodes", []) + node.get("floating_nodes", []):
            self._walk_sway_tree(child, windows)

    # ─── Window Finding ───

    async def find_window(self, query: str) -> dict[str, Any] | None:
        """Find a window by title or app name (fuzzy match)."""
        windows = await self.list_windows()
        query_lower = query.lower()

        # Exact title match first
        for w in windows:
            if query_lower == w.get("title", "").lower():
                return w

        # Substring match in title
        for w in windows:
            if query_lower in w.get("title", "").lower():
                return w

        # Substring match in app name
        for w in windows:
            if query_lower in w.get("app", "").lower():
                return w

        return None

    # ─── Window Focus ───

    async def focus(self, query: str) -> dict[str, Any]:
        """Bring a window to the foreground by title or app name."""
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        if SYSTEM == "Windows":
            cmd = (
                f'powershell -Command "'
                f"$p = Get-Process -Id {window['pid']}; "
                f"$sig = '[DllImport(\\\"user32.dll\\\")] public static extern bool SetForegroundWindow(IntPtr hWnd);'; "
                f"$type = Add-Type -MemberDefinition $sig -Name Win32Focus -PassThru; "
                f'$type::SetForegroundWindow($p.MainWindowHandle)"'
            )
        elif SYSTEM == "Darwin":
            app = window.get("app", query)
            cmd = f'osascript -e \'tell application "{app}" to activate\''
        else:
            wid = window.get("id", "")
            if wid and shutil.which("swaymsg"):
                cmd = f'swaymsg "[con_id={wid}] focus"'
            elif wid and shutil.which("wmctrl"):
                cmd = f"wmctrl -i -a {wid}"
            elif wid and shutil.which("xdotool"):
                cmd = f"xdotool windowactivate {wid}"
            else:
                return {"success": False, "error": "No window manager tool available"}

        _, err, code = await self._run(cmd)
        return {
            "success": code == 0,
            "action": "focus",
            "window": window,
            "error": err if code != 0 else None,
        }

    # ─── Window Close ───

    async def close(self, query: str) -> dict[str, Any]:
        """Close a window by title or app name."""
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        if SYSTEM == "Windows":
            cmd = f'powershell -Command "Stop-Process -Id {window["pid"]} -Force"'
        elif SYSTEM == "Darwin":
            app = window.get("app", query)
            cmd = f'osascript -e \'tell application "{app}" to quit\''
        else:
            wid = window.get("id", "")
            if wid and shutil.which("swaymsg"):
                cmd = f'swaymsg "[con_id={wid}] kill"'
            elif wid and shutil.which("wmctrl"):
                cmd = f"wmctrl -i -c {wid}"
            elif wid and shutil.which("xdotool"):
                cmd = f"xdotool windowclose {wid}"
            else:
                return {"success": False, "error": "No window manager tool available"}

        _, err, code = await self._run(cmd)
        return {"success": code == 0, "action": "close", "window": window}

    # ─── Window Minimize / Maximize ───

    async def minimize(self, query: str) -> dict[str, Any]:
        """Minimize a window."""
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        if SYSTEM == "Windows":
            cmd = (
                f'powershell -Command "'
                f"$sig = '[DllImport(\\\"user32.dll\\\")] public static extern bool ShowWindow(IntPtr h, int c);'; "
                f"$t = Add-Type -MemberDefinition $sig -Name Win32Min -PassThru; "
                f"$p = Get-Process -Id {window['pid']}; "
                f'$t::ShowWindow($p.MainWindowHandle, 6)"'
            )
        elif SYSTEM == "Darwin":
            app = window.get("app", query)
            cmd = f'osascript -e \'tell application "System Events" to set miniaturized of window 1 of process "{app}" to true\''
        else:
            wid = window.get("id", "")
            cmd = f"xdotool windowminimize {wid}" if shutil.which("xdotool") else ""

        if not cmd:
            return {"success": False, "error": "Cannot minimize on this platform"}

        _, err, code = await self._run(cmd)
        return {"success": code == 0, "action": "minimize", "window": window}

    async def maximize(self, query: str) -> dict[str, Any]:
        """Maximize a window."""
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        if SYSTEM == "Windows":
            cmd = (
                f'powershell -Command "'
                f"$sig = '[DllImport(\\\"user32.dll\\\")] public static extern bool ShowWindow(IntPtr h, int c);'; "
                f"$t = Add-Type -MemberDefinition $sig -Name Win32Max -PassThru; "
                f"$p = Get-Process -Id {window['pid']}; "
                f'$t::ShowWindow($p.MainWindowHandle, 3)"'
            )
        elif SYSTEM == "Darwin":
            app = window.get("app", query)
            cmd = f'osascript -e \'tell application "System Events" to set value of attribute "AXFullScreen" of window 1 of process "{app}" to true\''
        else:
            wid = window.get("id", "")
            if shutil.which("wmctrl"):
                cmd = f"wmctrl -i -r {wid} -b add,maximized_vert,maximized_horz"
            elif shutil.which("xdotool"):
                cmd = f"xdotool windowsize {wid} 100% 100%"
            else:
                return {"success": False, "error": "No window manager tool available"}

        _, err, code = await self._run(cmd)
        return {"success": code == 0, "action": "maximize", "window": window}

    # ─── Window Positioning ───

    async def move(self, query: str, x: int, y: int) -> dict[str, Any]:
        """Move a window to (x, y)."""
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        if SYSTEM == "Windows":
            cmd = (
                f'powershell -Command "'
                f"$sig = '[DllImport(\\\"user32.dll\\\")] public static extern bool MoveWindow(IntPtr h,int x,int y,int w,int h2,bool r);'; "
                f"$t = Add-Type -MemberDefinition $sig -Name Win32Move -PassThru; "
                f"$p = Get-Process -Id {window['pid']}; "
                f'$t::MoveWindow($p.MainWindowHandle, {x}, {y}, 800, 600, $true)"'
            )
        elif SYSTEM == "Darwin":
            app = window.get("app", query)
            cmd = f'osascript -e \'tell application "System Events" to set position of window 1 of process "{app}" to {{{x}, {y}}}\''
        else:
            wid = window.get("id", "")
            if shutil.which("wmctrl"):
                cmd = f"wmctrl -i -r {wid} -e 0,{x},{y},-1,-1"
            elif shutil.which("xdotool"):
                cmd = f"xdotool windowmove {wid} {x} {y}"
            else:
                return {"success": False, "error": "No window manager tool available"}

        _, err, code = await self._run(cmd)
        return {"success": code == 0, "action": "move", "window": window, "position": (x, y)}

    async def resize(self, query: str, width: int, height: int) -> dict[str, Any]:
        """Resize a window."""
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        if SYSTEM == "Windows":
            cmd = (
                f'powershell -Command "'
                f"$sig = '[DllImport(\\\"user32.dll\\\")] public static extern bool MoveWindow(IntPtr h,int x,int y,int w,int h2,bool r);'; "
                f"$t = Add-Type -MemberDefinition $sig -Name Win32Resize -PassThru; "
                f"$p = Get-Process -Id {window['pid']}; "
                f'$t::MoveWindow($p.MainWindowHandle, 0, 0, {width}, {height}, $true)"'
            )
        elif SYSTEM == "Darwin":
            app = window.get("app", query)
            cmd = f'osascript -e \'tell application "System Events" to set size of window 1 of process "{app}" to {{{width}, {height}}}\''
        else:
            wid = window.get("id", "")
            if shutil.which("wmctrl"):
                cmd = f"wmctrl -i -r {wid} -e 0,-1,-1,{width},{height}"
            elif shutil.which("xdotool"):
                cmd = f"xdotool windowsize {wid} {width} {height}"
            else:
                return {"success": False, "error": "No window manager tool available"}

        _, err, code = await self._run(cmd)
        return {"success": code == 0, "action": "resize", "window": window, "size": (width, height)}

    # ─── Smart Snapping ───

    async def snap_left(self, query: str) -> dict[str, Any]:
        """Snap window to the left half of the screen."""
        sw, sh = await self._get_screen_size()
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        await self.focus(query)
        await asyncio.sleep(0.1)

        # Use move + resize for precise control
        result = await self._snap_to(window, 0, 0, sw // 2, sh)
        result["action"] = "snap_left"
        return result

    async def snap_right(self, query: str) -> dict[str, Any]:
        """Snap window to the right half of the screen."""
        sw, sh = await self._get_screen_size()
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        await self.focus(query)
        await asyncio.sleep(0.1)

        result = await self._snap_to(window, sw // 2, 0, sw // 2, sh)
        result["action"] = "snap_right"
        return result

    async def snap_top(self, query: str) -> dict[str, Any]:
        """Snap window to the top half."""
        sw, sh = await self._get_screen_size()
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        result = await self._snap_to(window, 0, 0, sw, sh // 2)
        result["action"] = "snap_top"
        return result

    async def snap_bottom(self, query: str) -> dict[str, Any]:
        """Snap window to the bottom half."""
        sw, sh = await self._get_screen_size()
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        result = await self._snap_to(window, 0, sh // 2, sw, sh // 2)
        result["action"] = "snap_bottom"
        return result

    async def snap_quarter(self, query: str, position: str = "top_left") -> dict[str, Any]:
        """Snap window to a quarter of the screen.

        position: top_left, top_right, bottom_left, bottom_right
        """
        sw, sh = await self._get_screen_size()
        window = await self.find_window(query)
        if not window:
            return {"success": False, "error": f"Window not found: {query}"}

        hw, hh = sw // 2, sh // 2
        positions = {
            "top_left": (0, 0, hw, hh),
            "top_right": (hw, 0, hw, hh),
            "bottom_left": (0, hh, hw, hh),
            "bottom_right": (hw, hh, hw, hh),
        }

        coords = positions.get(position, positions["top_left"])
        result = await self._snap_to(window, *coords)
        result["action"] = f"snap_{position}"
        return result

    async def _snap_to(
        self, window: dict, x: int, y: int, w: int, h: int
    ) -> dict[str, Any]:
        """Move and resize a window to exact position."""
        wid = window.get("id", "")

        if SYSTEM == "Windows":
            cmd = (
                f'powershell -Command "'
                f"$sig = '[DllImport(\\\"user32.dll\\\")] public static extern bool MoveWindow(IntPtr h,int x,int y,int w,int h2,bool r);'; "
                f"$t = Add-Type -MemberDefinition $sig -Name Win32Snap -PassThru; "
                f"$p = Get-Process -Id {window.get('pid', 0)}; "
                f'$t::MoveWindow($p.MainWindowHandle, {x}, {y}, {w}, {h}, $true)"'
            )
        elif SYSTEM == "Darwin":
            app = window.get("app", "")
            cmd = (
                f'osascript -e \'tell application "System Events" to tell process "{app}" to '
                f"set position of window 1 to {{{x}, {y}}}' "
                f'-e \'tell application "System Events" to tell process "{app}" to '
                f"set size of window 1 to {{{w}, {h}}}'"
            )
        else:
            if shutil.which("swaymsg") and wid:
                # Sway: use floating + move/resize
                cmd = (
                    f'swaymsg "[con_id={wid}] floating enable, '
                    f'move position {x} {y}, '
                    f'resize set {w} {h}"'
                )
            elif shutil.which("wmctrl") and wid:
                cmd = f"wmctrl -i -r {wid} -e 0,{x},{y},{w},{h}"
            elif shutil.which("xdotool") and wid:
                cmd = f"xdotool windowmove {wid} {x} {y} && xdotool windowsize {wid} {w} {h}"
            else:
                return {"success": False, "error": "No window manager tool available"}

        _, err, code = await self._run(cmd)
        return {
            "success": code == 0,
            "window": window,
            "bounds": {"x": x, "y": y, "width": w, "height": h},
        }

    # ─── Tiling ───

    async def tile_windows(self, queries: list[str]) -> dict[str, Any]:
        """Tile multiple windows evenly across the screen.

        Arranges 2 windows side-by-side, 3 in thirds, 4 in quadrants, etc.
        """
        sw, sh = await self._get_screen_size()
        n = len(queries)

        if n == 0:
            return {"success": False, "error": "No windows specified"}

        results = []

        if n == 1:
            results.append(await self.maximize(queries[0]))
        elif n == 2:
            results.append(await self.snap_left(queries[0]))
            results.append(await self.snap_right(queries[1]))
        elif n == 3:
            third = sw // 3
            for i, q in enumerate(queries):
                w = await self.find_window(q)
                if w:
                    r = await self._snap_to(w, third * i, 0, third, sh)
                    results.append(r)
        elif n == 4:
            positions = ["top_left", "top_right", "bottom_left", "bottom_right"]
            for q, pos in zip(queries, positions):
                results.append(await self.snap_quarter(q, pos))
        else:
            # Grid layout for 5+
            cols = 3 if n > 4 else 2
            rows = (n + cols - 1) // cols
            cw = sw // cols
            ch = sh // rows
            for i, q in enumerate(queries):
                row, col = divmod(i, cols)
                w = await self.find_window(q)
                if w:
                    r = await self._snap_to(w, col * cw, row * ch, cw, ch)
                    results.append(r)

        return {
            "action": "tile",
            "count": n,
            "results": results,
            "success": all(r.get("success", False) for r in results),
        }

    # ─── Convenience ───

    async def switch_to(self, query: str) -> dict[str, Any]:
        """Switch to a window (alias for focus with a small delay)."""
        result = await self.focus(query)
        await asyncio.sleep(0.3)
        return result

    async def get_active_window(self) -> dict[str, Any]:
        """Get info about the currently focused window."""
        if SYSTEM == "Windows":
            cmd = (
                'powershell -Command "'
                "Add-Type -TypeDefinition 'using System; using System.Runtime.InteropServices; "
                "public class Win32 { [DllImport(\"user32.dll\")] public static extern IntPtr GetForegroundWindow(); "
                "[DllImport(\"user32.dll\")] public static extern int GetWindowText(IntPtr h, System.Text.StringBuilder t, int c); "
                "[DllImport(\"user32.dll\")] public static extern uint GetWindowThreadProcessId(IntPtr h, out uint pid); }'; "
                "$h = [Win32]::GetForegroundWindow(); "
                "$sb = New-Object System.Text.StringBuilder 256; "
                "[Win32]::GetWindowText($h, $sb, 256); "
                "$pid = 0; [Win32]::GetWindowThreadProcessId($h, [ref]$pid); "
                'Write-Output \"$($sb.ToString())|$pid\"'
                '"'
            )
            out, _, _ = await self._run(cmd)
            parts = out.split("|")
            return {"title": parts[0] if parts else "", "pid": int(parts[1]) if len(parts) > 1 and parts[1].isdigit() else 0}
        elif SYSTEM == "Darwin":
            cmd = 'osascript -e \'tell application "System Events" to get name of first process whose frontmost is true\''
            out, _, _ = await self._run(cmd)
            return {"app": out.strip(), "title": out.strip()}
        else:
            # Try swaymsg first (Wayland/Sway)
            if shutil.which("swaymsg"):
                try:
                    import json as _json
                    out, _, rc = await self._run("swaymsg -t get_tree --raw")
                    if rc == 0:
                        tree = _json.loads(out)
                        focused = self._find_sway_focused(tree)
                        if focused:
                            return {
                                "id": str(focused.get("id", "")),
                                "title": focused.get("name", ""),
                                "app": focused.get("app_id", ""),
                            }
                except Exception:
                    pass
            if shutil.which("xdotool"):
                wid_out, _, _ = await self._run("xdotool getactivewindow")
                name_out, _, _ = await self._run(f"xdotool getwindowname {wid_out.strip()}")
                return {"id": wid_out.strip(), "title": name_out.strip()}
            return {"error": "No window tool available (install xdotool or swaymsg)"}
